cached_aws_sandbox_creds = None
cached_provider_class = None


class Boto3ClientProvider(object):
    name = "boto3"

    @staticmethod
    def get_client(
        module, with_error=False, role_arn=None, session_vars=None, client_params=None
    ):
        from metaflow.exception import MetaflowException
        from metaflow.metaflow_config import (
            AWS_SANDBOX_ENABLED,
            AWS_SANDBOX_STS_ENDPOINT_URL,
            AWS_SANDBOX_API_KEY,
            S3_CLIENT_RETRY_CONFIG,
        )

        if session_vars is None:
            session_vars = {}

        if client_params is None:
            client_params = {}

        import requests

        try:
            import boto3
            import botocore
            from botocore.exceptions import ClientError
            from botocore.config import Config
        except (NameError, ImportError):
            raise MetaflowException(
                "Could not import module 'boto3'. Install boto3 first."
            )

        # Convert dictionary config to Config object if needed
        if "config" in client_params and not isinstance(
            client_params["config"], Config
        ):
            client_params["config"] = Config(**client_params["config"])

        if module == "s3" and (
            "config" not in client_params or client_params["config"].retries is None
        ):
            # do not set anything if the user has already set something
            config = client_params.get("config", Config())
            config.retries = S3_CLIENT_RETRY_CONFIG
            client_params["config"] = config

        if AWS_SANDBOX_ENABLED:
            # role is ignored in the sandbox
            global cached_aws_sandbox_creds
            if cached_aws_sandbox_creds is None:
                # authenticate using STS
                url = "%s/auth/token" % AWS_SANDBOX_STS_ENDPOINT_URL
                headers = {"x-api-key": AWS_SANDBOX_API_KEY}
                try:
                    r = requests.get(url, headers=headers)
                    r.raise_for_status()
                    cached_aws_sandbox_creds = r.json()
                except requests.exceptions.HTTPError as e:
                    raise MetaflowException(repr(e))
            if with_error:
                return (
                    boto3.session.Session(**cached_aws_sandbox_creds).client(
                        module, **client_params
                    ),
                    ClientError,
                )
            return boto3.session.Session(**cached_aws_sandbox_creds).client(
                module, **client_params
            )
        session = boto3.session.Session()
        if role_arn:
            fetcher = botocore.credentials.AssumeRoleCredentialFetcher(
                client_creator=session._session.create_client,
                source_credentials=session._session.get_credentials(),
                role_arn=role_arn,
                extra_args={},
            )
            creds = botocore.credentials.DeferredRefreshableCredentials(
                method="assume-role", refresh_using=fetcher.fetch_credentials
            )
            botocore_session = botocore.session.Session(session_vars=session_vars)
            botocore_session._credentials = creds
            session = boto3.session.Session(botocore_session=botocore_session)
        if with_error:
            return session.client(module, **client_params), ClientError
        return session.client(module, **client_params)


def get_aws_client(
    module, with_error=False, role_arn=None, session_vars=None, client_params=None
):
    global cached_provider_class
    if cached_provider_class is None:
        from metaflow.metaflow_config import DEFAULT_AWS_CLIENT_PROVIDER
        from metaflow.plugins import AWS_CLIENT_PROVIDERS

        for p in AWS_CLIENT_PROVIDERS:
            if p.name == DEFAULT_AWS_CLIENT_PROVIDER:
                cached_provider_class = p
                break
        else:
            raise ValueError(
                "Cannot find AWS Client provider %s" % DEFAULT_AWS_CLIENT_PROVIDER
            )
    return cached_provider_class.get_client(
        module,
        with_error,
        role_arn=role_arn,
        session_vars=session_vars,
        client_params=client_params,
    )



import os
import platform
import re
import sys
import tempfile

from metaflow.decorators import FlowDecorator, StepDecorator
from metaflow.metadata_provider import MetaDatum
from metaflow.metaflow_environment import InvalidEnvironmentException
from metaflow.packaging_sys import ContentType


class CondaStepDecorator(StepDecorator):
    """
    Specifies the Conda environment for the step.

    Information in this decorator will augment any
    attributes set in the `@conda_base` flow-level decorator. Hence,
    you can use `@conda_base` to set packages required by all
    steps and use `@conda` to specify step-specific overrides.

    Parameters
    ----------
    packages : Dict[str, str], default {}
        Packages to use for this step. The key is the name of the package
        and the value is the version to use.
    libraries : Dict[str, str], default {}
        Supported for backward compatibility. When used with packages, packages will take precedence.
    python : str, optional, default None
        Version of Python to use, e.g. '3.7.4'. A default value of None implies
        that the version used will correspond to the version of the Python interpreter used to start the run.
    disabled : bool, default False
        If set to True, disables @conda.
    """

    name = "conda"
    defaults = {
        "packages": {},
        "libraries": {},  # Deprecated! Use packages going forward
        "python": None,
        "disabled": None,
    }

    _metaflow_home = None
    _addl_env_vars = None

    # To define conda channels for the whole solve, users can specify
    # CONDA_CHANNELS in their environment. For pinning specific packages to specific
    # conda channels, users can specify channel::package as the package name.

    def __init__(self, attributes=None, statically_defined=False, inserted_by=None):
        self._attributes_with_user_values = (
            set(attributes.keys()) if attributes is not None else set()
        )

        super(CondaStepDecorator, self).__init__(
            attributes, statically_defined, inserted_by
        )

    def init(self):
        # Support legacy 'libraries=' attribute for the decorator.
        self.attributes["packages"] = {
            **self.attributes["libraries"],
            **self.attributes["packages"],
        }
        # Keep because otherwise make_decorator_spec will fail
        self.attributes["libraries"] = {}
        if self.attributes["packages"]:
            self._attributes_with_user_values.add("packages")

    def is_attribute_user_defined(self, name):
        return name in self._attributes_with_user_values

    def step_init(self, flow, graph, step, decos, environment, flow_datastore, logger):
        # The init_environment hook for Environment creates the relevant virtual
        # environments. The step_init hook sets up the relevant state for that hook to
        # do it's magic.

        self.flow = flow
        self.step = step
        self.environment = environment
        self.datastore = flow_datastore

        # Support flow-level decorator.
        if "conda_base" in self.flow._flow_decorators:
            conda_base = self.flow._flow_decorators["conda_base"][0]
            super_attributes = conda_base.attributes
            self.attributes["packages"] = {
                **super_attributes["packages"],
                **self.attributes["packages"],
            }
            self._attributes_with_user_values.update(
                conda_base._attributes_with_user_values
            )

            self.attributes["python"] = (
                self.attributes["python"] or super_attributes["python"]
            )
            self.attributes["disabled"] = (
                self.attributes["disabled"]
                if self.attributes["disabled"] is not None
                else super_attributes["disabled"]
            )

        # Set default for `disabled` argument.
        if not self.attributes["disabled"]:
            self.attributes["disabled"] = False
        # Set Python interpreter to user's Python if necessary.
        if not self.attributes["python"]:
            self.attributes["python"] = platform.python_version()  # CPython!

        # @conda uses a conda environment to create a virtual environment.
        # The conda environment can be created through micromamba.
        _supported_virtual_envs = ["conda"]

        # To placate people who don't want to see a shred of conda in UX, we symlink
        # --environment=pypi to --environment=conda
        _supported_virtual_envs.extend(["pypi"])

        # TODO: Hardcoded for now to support the fast bakery environment.
        # We should introduce a more robust mechanism for appending supported environments, for example from within extensions.
        _supported_virtual_envs.extend(["fast-bakery"])

        # The --environment= requirement ensures that valid virtual environments are
        # created for every step to execute it, greatly simplifying the @conda
        # implementation.
        if environment.TYPE not in _supported_virtual_envs:
            raise InvalidEnvironmentException(
                "@%s decorator requires %s"
                % (
                    self.name,
                    " or ".join(
                        ["--environment=%s" % env for env in _supported_virtual_envs]
                    ),
                )
            )

        # At this point, the list of 32 bit instance types is shrinking quite rapidly.
        # We can worry about supporting them when there is a need.

        # TODO: This code snippet can be done away with by altering the constructor of
        #       MetaflowEnvironment. A good first-task exercise.
        # Avoid circular import
        from metaflow.plugins.datastores.local_storage import LocalStorage

        environment.set_local_root(LocalStorage.get_datastore_root_from_config(logger))

        self.disabled = self.environment.is_disabled(
            next(step for step in self.flow if step.name == self.step)
        )

    def runtime_init(self, flow, graph, package, run_id):
        if self.disabled:
            return
        # We need to make all the code package available to the user code in
        # a temporary directory which will be added to the PYTHONPATH.
        if self.__class__._metaflow_home is None:
            # Do this ONCE per flow
            self.__class__._metaflow_home = tempfile.TemporaryDirectory(dir="/tmp")
            package.extract_into(
                self.__class__._metaflow_home.name, ContentType.ALL_CONTENT
            )
            self.__class__._addl_env_vars = package.get_post_extract_env_vars(
                package.package_metadata, self.__class__._metaflow_home.name
            )

        # # Also install any environment escape overrides directly here to enable
        # # the escape to work even in non metaflow-created subprocesses
        # from ..env_escape import generate_trampolines
        # generate_trampolines(self.metaflow_dir.name)

    def runtime_task_created(
        self, task_datastore, task_id, split_index, input_paths, is_cloned, ubf_context
    ):
        if self.disabled:
            return
        self.interpreter = (
            self.environment.interpreter(self.step)
            if not any(
                decorator.name
                in ["batch", "kubernetes", "nvidia", "snowpark", "slurm", "nvct"]
                for decorator in next(
                    step for step in self.flow if step.name == self.step
                ).decorators
            )
            else None
        )

    def task_pre_step(
        self,
        step_name,
        task_datastore,
        meta,
        run_id,
        task_id,
        flow,
        graph,
        retry_count,
        max_retries,
        ubf_context,
        inputs,
    ):
        if self.disabled:
            return
        # Add Python interpreter's parent to the path to ensure that any non-pythonic
        # dependencies in the virtual environment are visible to the user code.
        # sys.executable points to the Python interpreter in the virtual environment
        # since we are already inside the task context.
        os.environ["PATH"] = os.pathsep.join(
            filter(
                None,
                (
                    os.path.dirname(os.path.realpath(sys.executable)),
                    os.environ.get("PATH"),
                ),
            )
        )

        # Infer environment prefix from Python interpreter
        match = re.search(
            r"(?:.*\/)(metaflow\/[^/]+\/[^/]+)(?=\/bin\/python)", sys.executable
        )
        if match:
            meta.register_metadata(
                run_id,
                step_name,
                task_id,
                [
                    MetaDatum(
                        field="conda_env_prefix",
                        value=match.group(1),
                        type="conda_env_prefix",
                        tags=["attempt_id:{0}".format(retry_count)],
                    )
                ],
            )

    def runtime_step_cli(
        self, cli_args, retry_count, max_user_code_retries, ubf_context
    ):
        if self.disabled:
            return
        # Ensure local installation of Metaflow is visible to user code
        python_path = self.__class__._metaflow_home.name
        addl_env_vars = {}
        if self.__class__._addl_env_vars is not None:
            for key, value in self.__class__._addl_env_vars.items():
                if key == "PYTHONPATH":
                    addl_env_vars[key] = os.pathsep.join([value, python_path])
                else:
                    addl_env_vars[key] = value
        cli_args.env.update(addl_env_vars)
        if self.interpreter:
            # https://github.com/conda/conda/issues/7707
            # Also ref - https://github.com/Netflix/metaflow/pull/178
            cli_args.env["PYTHONNOUSERSITE"] = "1"
            # The executable is already in place for the user code to execute against
            cli_args.entrypoint[0] = self.interpreter

    def runtime_finished(self, exception):
        if self.disabled:
            return
        if self.__class__._metaflow_home is not None:
            self.__class__._metaflow_home.cleanup()
            self.__class__._metaflow_home = None


class CondaFlowDecorator(FlowDecorator):
    """
    Specifies the Conda environment for all steps of the flow.

    Use `@conda_base` to set common libraries required by all
    steps and use `@conda` to specify step-specific additions.

    Parameters
    ----------
    packages : Dict[str, str], default {}
        Packages to use for this flow. The key is the name of the package
        and the value is the version to use.
    libraries : Dict[str, str], default {}
        Supported for backward compatibility. When used with packages, packages will take precedence.
    python : str, optional, default None
        Version of Python to use, e.g. '3.7.4'. A default value of None implies
        that the version used will correspond to the version of the Python interpreter used to start the run.
    disabled : bool, default False
        If set to True, disables Conda.
    """

    # TODO: Migrate conda_base keyword to conda for simplicity.
    name = "conda_base"
    defaults = {
        "packages": {},
        "libraries": {},  # Deprecated! Use packages going forward.
        "python": None,
        "disabled": None,
    }

    def __init__(self, attributes=None, statically_defined=False, inserted_by=None):
        self._attributes_with_user_values = (
            set(attributes.keys()) if attributes is not None else set()
        )

        super(CondaFlowDecorator, self).__init__(
            attributes, statically_defined, inserted_by
        )

    def init(self):
        # Support legacy 'libraries=' attribute for the decorator.
        self.attributes["packages"] = {
            **self.attributes["libraries"],
            **self.attributes["packages"],
        }
        # Keep because otherwise make_decorator_spec will fail
        self.attributes["libraries"] = {}
        if self.attributes["python"]:
            self.attributes["python"] = str(self.attributes["python"])

    def is_attribute_user_defined(self, name):
        return name in self._attributes_with_user_values

    def flow_init(
        self, flow, graph, environment, flow_datastore, metadata, logger, echo, options
    ):
        # NOTE: Important for extensions implementing custom virtual environments.
        # Without this steps will not have an implicit conda step decorator on them unless the environment adds one in its decospecs.
        from metaflow import decorators

        decorators._attach_decorators(flow, ["conda"])

        # @conda uses a conda environment to create a virtual environment.
        # The conda environment can be created through micromamba.
        _supported_virtual_envs = ["conda"]

        # To placate people who don't want to see a shred of conda in UX, we symlink
        # --environment=pypi to --environment=conda
        _supported_virtual_envs.extend(["pypi"])

        # TODO: Hardcoded for now to support the fast bakery environment.
        # We should introduce a more robust mechanism for appending supported environments, for example from within extensions.
        _supported_virtual_envs.extend(["fast-bakery"])

        # The --environment= requirement ensures that valid virtual environments are
        # created for every step to execute it, greatly simplifying the @conda
        # implementation.
        if environment.TYPE not in _supported_virtual_envs:
            raise InvalidEnvironmentException(
                "@%s decorator requires %s"
                % (
                    self.name,
                    " or ".join(
                        ["--environment=%s" % env for env in _supported_virtual_envs]
                    ),
                )
            )



import json
import os
import sys
import time
import urllib
import uuid
from datetime import datetime

from metaflow.exception import MetaflowException
from metaflow.metaflow_config import (
    ARGO_EVENTS_WEBHOOK_AUTH,
    ARGO_EVENTS_WEBHOOK_URL,
    SERVICE_HEADERS,
)


class ArgoEventException(MetaflowException):
    headline = "Argo Event Exception"


class ArgoEvent(object):
    """
    ArgoEvent is a small event, a message, that can be published to Argo Workflows. The
    event will eventually start all flows which have been previously deployed with `@trigger`
    to wait for this particular named event.

    Parameters
    ----------
    name : str,
        Name of the event
    url : str, optional
        Override the event endpoint from `ARGO_EVENTS_WEBHOOK_URL`.
    payload : Dict, optional
        A set of key-value pairs delivered in this event. Used to set parameters of triggered flows.
    """

    def __init__(
        self, name, url=ARGO_EVENTS_WEBHOOK_URL, payload=None, access_token=None
    ):
        # TODO: Introduce support for NATS
        self._name = name
        self._url = url
        self._payload = payload or {}
        self._access_token = access_token

    def add_to_payload(self, key, value):
        """
        Add a key-value pair in the payload. This is typically used to set parameters
        of triggered flows. Often, `key` is the parameter name you want to set to
        `value`. Overrides any existing value of `key`.

        Parameters
        ----------
        key : str
            Key
        value : str
            Value
        """

        self._payload[key] = str(value)
        return self

    def safe_publish(self, payload=None, ignore_errors=True):
        """
        Publishes an event when called inside a deployed workflow. Outside a deployed workflow
        this function does nothing.

        Use this function inside flows to create events safely. As this function is a no-op
        for local runs, you can safely call it during local development without causing unintended
        side-effects. It takes effect only when deployed on Argo Workflows.

        Parameters
        ----------
        payload : dict
            Additional key-value pairs to add to the payload.
        ignore_errors : bool, default True
            If True, events are created on a best effort basis - errors are silently ignored.
        """

        return self.publish(payload=payload, force=False, ignore_errors=ignore_errors)

    def publish(self, payload=None, force=True, ignore_errors=True):
        """
        Publishes an event.

        Note that the function returns immediately after the event has been sent. It
        does not wait for flows to start, nor it guarantees that any flows will start.

        Parameters
        ----------
        payload : dict
            Additional key-value pairs to add to the payload.
        ignore_errors : bool, default True
            If True, events are created on a best effort basis - errors are silently ignored.
        """

        if payload == None:
            payload = {}
        # Publish event iff forced or running on Argo Workflows
        if force or os.environ.get("ARGO_WORKFLOW_TEMPLATE"):
            try:
                headers = {}
                if self._access_token:
                    # TODO: Test with bearer tokens
                    headers = {"Authorization": "Bearer {}".format(self._access_token)}
                if ARGO_EVENTS_WEBHOOK_AUTH == "service":
                    headers.update(SERVICE_HEADERS)
                # TODO: do we need to worry about certs?

                # Use urllib to avoid introducing any dependency in Metaflow
                data = {
                    "name": self._name,
                    "payload": {
                        # Add default fields here...
                        "name": self._name,
                        "id": str(uuid.uuid4()),
                        "timestamp": int(time.time()),
                        "utc_date": datetime.utcnow().strftime("%Y%m%d"),
                        "generated-by-metaflow": True,
                        **self._payload,
                        **payload,
                    },
                }
                request = urllib.request.Request(
                    self._url,
                    method="POST",
                    headers={"Content-Type": "application/json", **headers},
                    data=json.dumps(data).encode("utf-8"),
                )
                retries = 3
                backoff_factor = 2

                for i in range(retries):
                    try:
                        urllib.request.urlopen(request, timeout=10.0)
                        print(
                            "Argo Event (%s) published." % self._name, file=sys.stderr
                        )
                        return data["payload"]["id"]
                    except urllib.error.HTTPError as e:
                        # TODO: Retry retryable HTTP error codes
                        raise e
                    except urllib.error.URLError as e:
                        if i == retries - 1:
                            raise e
                        else:
                            time.sleep(backoff_factor**i)
            except Exception as e:
                msg = "Unable to publish Argo Event (%s): %s" % (self._name, e)
                if ignore_errors:
                    print(msg, file=sys.stderr)
                else:
                    raise ArgoEventException(msg)
        else:
            msg = (
                "Argo Event (%s) was not published. Use "
                + "ArgoEvent(...).publish(...) "
                + "to force publish."
            ) % self._name

            if ignore_errors:
                print(msg, file=sys.stderr)
            else:
                raise ArgoEventException(msg)



import os
import shutil
import sys
from subprocess import PIPE, CompletedProcess, run
from tempfile import TemporaryDirectory
from typing import Any, Callable, List, Mapping, Optional, cast

from metaflow import Run
from metaflow._vendor import click
from metaflow.cli import echo_always


@click.group()
def cli():
    pass


@cli.group(help="Access, compare, and manage code associated with Metaflow runs.")
def code():
    pass


def echo(line: str) -> None:
    echo_always(line, err=True, fg="magenta")


def extract_code_package(runspec: str) -> TemporaryDirectory:
    try:
        mf_run = Run(runspec, _namespace_check=False)
        echo(f"✅  Run *{runspec}* found, downloading code..")
    except Exception as e:
        echo(f"❌  Run **{runspec}** not found")
        raise e

    if mf_run.code is None:
        echo(
            f"❌  Run **{runspec}** doesn't have a code package. Maybe it's a local run?"
        )
        raise RuntimeError("no code package found")

    return mf_run.code.extract()


def perform_diff(
    source_dir: str,
    target_dir: Optional[str] = None,
    output: bool = False,
    **kwargs: Mapping[str, Any],
) -> Optional[List[str]]:
    if target_dir is None:
        target_dir = os.getcwd()

    diffs = []
    for dirpath, dirnames, filenames in os.walk(source_dir, followlinks=True):
        for fname in filenames:
            # NOTE: the paths below need to be set up carefully
            # for the `patch` command to work. Better not to touch
            # the directories below. If you must, test that patches
            # work after your changes.
            #
            # target_file is the git repo in the current working directory
            rel = os.path.relpath(dirpath, source_dir)
            target_file = os.path.join(rel, fname)
            # source_file is the run file loaded in a tmp directory
            source_file = os.path.join(dirpath, fname)

            if sys.stdout.isatty() and not output:
                color = ["--color"]
            else:
                color = ["--no-color"]

            if os.path.exists(os.path.join(target_dir, target_file)):
                cmd = (
                    ["git", "diff", "--no-index", "--exit-code"]
                    + color
                    + [
                        target_file,
                        source_file,
                    ]
                )
                result: CompletedProcess = run(
                    cmd, text=True, stdout=PIPE, cwd=target_dir
                )
                if result.returncode == 0:
                    if not output:
                        echo(f"✅  {target_file} is identical, skipping")
                    continue

                if output:
                    diffs.append(result.stdout)
                else:
                    run(["less", "-R"], input=result.stdout, text=True)
            else:
                if not output:
                    echo(f"❗  {target_file} not in the target directory, skipping")
    return diffs if output else None


def run_op(
    runspec: str, op: Callable[..., Optional[List[str]]], **op_args: Mapping[str, Any]
) -> Optional[List[str]]:
    tmp = None
    try:
        tmp = extract_code_package(runspec)
        return op(tmp.name, **op_args)
    finally:
        if tmp and os.path.exists(tmp.name):
            shutil.rmtree(tmp.name)


def run_op_diff_runs(
    source_run_pathspec: str, target_run_pathspec: str, **op_args: Mapping[str, Any]
) -> Optional[List[str]]:
    source_tmp = None
    target_tmp = None
    try:
        source_tmp = extract_code_package(source_run_pathspec)
        target_tmp = extract_code_package(target_run_pathspec)
        return perform_diff(source_tmp.name, target_tmp.name, **op_args)
    finally:
        for d in [source_tmp, target_tmp]:
            if d and os.path.exists(d.name):
                shutil.rmtree(d.name)


def op_diff(tmpdir: str, **kwargs: Mapping[str, Any]) -> Optional[List[str]]:
    kwargs_dict = dict(kwargs)
    target_dir = cast(Optional[str], kwargs_dict.pop("target_dir", None))
    output: bool = bool(kwargs_dict.pop("output", False))
    op_args: Mapping[str, Any] = {**kwargs_dict}
    return perform_diff(tmpdir, target_dir=target_dir, output=output, **op_args)


def op_pull(tmpdir: str, dst: str, **op_args: Mapping[str, Any]) -> None:
    if os.path.exists(dst):
        echo(f"❌  Directory *{dst}* already exists")
    else:
        shutil.move(tmpdir, dst)
        echo(f"Code downloaded to *{dst}*")


def op_patch(tmpdir: str, dst: str, **kwargs: Mapping[str, Any]) -> None:
    diffs = perform_diff(tmpdir, output=True) or []
    with open(dst, "w", encoding="utf-8") as f:
        for out in diffs:
            out = out.replace(tmpdir, "/.")
            out = out.replace("+++ b/./", "+++ b/")
            out = out.replace("--- b/./", "--- b/")
            out = out.replace("--- a/./", "--- a/")
            out = out.replace("+++ a/./", "+++ a/")
            f.write(out)
    echo(f"Patch saved in *{dst}*")
    path = run(
        ["git", "rev-parse", "--show-prefix"], text=True, stdout=PIPE
    ).stdout.strip()
    if path:
        diropt = f" --directory={path.rstrip('/')}"
    else:
        diropt = ""
    echo("Apply the patch by running:")
    echo_always(
        f"git apply --verbose{diropt} {dst}", highlight=True, bold=True, err=True
    )


@code.command()
@click.argument("run_pathspec")
def diff(run_pathspec: str, **kwargs: Mapping[str, Any]) -> None:
    """
    Do a 'git diff' of the current directory and a Metaflow run.
    """
    _ = run_op(run_pathspec, op_diff, **kwargs)


@code.command()
@click.argument("source_run_pathspec")
@click.argument("target_run_pathspec")
def diff_runs(
    source_run_pathspec: str, target_run_pathspec: str, **kwargs: Mapping[str, Any]
) -> None:
    """
    Do a 'git diff' between two Metaflow runs.
    """
    _ = run_op_diff_runs(source_run_pathspec, target_run_pathspec, **kwargs)


@code.command()
@click.argument("run_pathspec")
@click.option(
    "--dir", help="Destination directory (default: {run_pathspec}_code)", default=None
)
def pull(
    run_pathspec: str, dir: Optional[str] = None, **kwargs: Mapping[str, Any]
) -> None:
    """
    Pull the code of a Metaflow run.
    """
    if dir is None:
        dir = run_pathspec.lower().replace("/", "_") + "_code"
    op_args: Mapping[str, Any] = {**kwargs, "dst": dir}
    run_op(run_pathspec, op_pull, **op_args)


@code.command()
@click.argument("run_pathspec")
@click.option(
    "--file_path",
    help="Patch file name. If not provided, defaults to a sanitized version of RUN_PATHSPEC "
    "with slashes replaced by underscores, plus '.patch'.",
    show_default=False,
)
@click.option(
    "--overwrite", is_flag=True, help="Overwrite the patch file if it exists."
)
def patch(
    run_pathspec: str,
    file_path: Optional[str] = None,
    overwrite: bool = False,
    **kwargs: Mapping[str, Any],
) -> None:
    """
    Create a patch by comparing current dir with a Metaflow run.
    """
    if file_path is None:
        file_path = run_pathspec.lower().replace("/", "_") + ".patch"
    if os.path.exists(file_path) and not overwrite:
        echo(f"File *{file_path}* already exists. To overwrite, specify --overwrite.")
        return
    op_args: Mapping[str, Any] = {**kwargs, "dst": file_path}
    run_op(run_pathspec, op_patch, **op_args)



import json
import os

from enum import IntEnum
from types import ModuleType
from typing import (
    Any,
    Dict,
    Generator,
    List,
    Optional,
    TYPE_CHECKING,
    Tuple,
    Type,
    Union,
)

from metaflow.packaging_sys.distribution_support import PackagedDistributionFinder


from .backend import PackagingBackend
from .tar_backend import TarPackagingBackend

from ..util import get_metaflow_root

MFCONTENT_MARKER = ".mf_install"

if TYPE_CHECKING:
    import metaflow.extension_support.metadata


class ContentType(IntEnum):
    USER_CONTENT = (
        0x1  # File being added is user code (ie: the directory with the flow file)
    )
    CODE_CONTENT = (
        0x2  # File being added is non-user code (libraries, metaflow itself, ...)
    )
    MODULE_CONTENT = 0x4  # File being added is a python module
    OTHER_CONTENT = 0x8  # File being added is a non-python file

    ALL_CONTENT = USER_CONTENT | CODE_CONTENT | MODULE_CONTENT | OTHER_CONTENT


class MetaflowCodeContent:
    """
    Base class for all Metaflow code packages (non user code).

    A Metaflow code package, at a minimum, contains:
      - a special INFO file (containing a bunch of metadata about the Metaflow environment)
      - a special CONFIG file (containing user configurations for the flow)

    Declare all other MetaflowCodeContent subclasses (versions) here to handle just the functions
    that are not implemented here. In a *separate* file, declare any other
    function for that specific version.

    NOTE: This file must remain as dependency-free as possible as it is loaded *very*
    early on. This is why you must decleare a *separate* class implementing what you want
    the Metaflow code package (non user) to do.
    """

    _cached_mfcontent_info = {}

    _mappings = {}

    @classmethod
    def get_info(cls) -> Optional[Dict[str, Any]]:
        """
        Get the content of the special INFO file on the local filesystem after
        the code package has been expanded.

        Returns
        -------
        Optional[Dict[str, Any]]
            The content of the INFO file -- None if there is no such file.
        """
        mfcontent_info = cls._extract_mfcontent_info()
        handling_cls = cls._get_mfcontent_class(mfcontent_info)
        return handling_cls.get_info_impl(mfcontent_info)

    @classmethod
    def get_config(cls) -> Optional[Dict[str, Any]]:
        """
        Get the content of the special CONFIG file on the local filesystem after
        the code package has been expanded.

        Returns
        -------
        Optional[Dict[str, Any]]
            The content of the CONFIG file -- None if there is no such file.
        """
        mfcontent_info = cls._extract_mfcontent_info()
        handling_cls = cls._get_mfcontent_class(mfcontent_info)
        return handling_cls.get_config_impl(mfcontent_info)

    @classmethod
    def get_filename(cls, filename: str, content_type: ContentType) -> Optional[str]:
        """
        Get the path to a file extracted from the archive. The filename is the filename
        passed in when creating the archive and content_type is the type of the content.

        This function will return the local path where the file can be found after
        the package has been extracted.

        Parameters
        ----------
        filename: str
            The name of the file on the filesystem.
        content_type: ContentType

        Returns
        -------
        str
            The path to the file on the local filesystem or None if not found.
        """
        mfcontent_info = cls._extract_mfcontent_info()
        handling_cls = cls._get_mfcontent_class(mfcontent_info)
        return handling_cls.get_filename_impl(mfcontent_info, filename, content_type)

    @classmethod
    def get_env_vars_for_packaged_metaflow(
        cls, dest_dir: str
    ) -> Optional[Dict[str, str]]:
        """
        Get the environment variables that are needed to run Metaflow when it is
        packaged. This is typically used to set the PYTHONPATH to include the
        directory where the Metaflow code package has been extracted.

        Returns
        -------
        Optional[Dict[str, str]]
            The environment variables that are needed to run Metaflow when it is
            packaged -- None if there are no such variables (not packaged for example)
        """
        mfcontent_info = cls._extract_mfcontent_info()
        if mfcontent_info is None:
            # No MFCONTENT_MARKER file found -- this is not a packaged Metaflow code
            # package so no environment variables to set.
            return None
        handling_cls = cls._get_mfcontent_class(mfcontent_info)
        return handling_cls.get_post_extract_env_vars_impl(dest_dir)

    @classmethod
    def get_archive_info(
        cls,
        archive: Any,
        packaging_backend: Type[PackagingBackend] = TarPackagingBackend,
    ) -> Optional[Dict[str, Any]]:
        """
        Get the content of the special INFO file in the archive.

        Returns
        -------
        Optional[Dict[str, Any]]
            The content of the INFO file -- None if there is no such file.
        """
        mfcontent_info = cls._extract_archive_mfcontent_info(archive, packaging_backend)
        handling_cls = cls._get_mfcontent_class(mfcontent_info)
        return handling_cls.get_archive_info_impl(
            mfcontent_info, archive, packaging_backend
        )

    @classmethod
    def get_archive_config(
        cls,
        archive: Any,
        packaging_backend: Type[PackagingBackend] = TarPackagingBackend,
    ) -> Optional[Dict[str, Any]]:
        """
        Get the content of the special CONFIG file in the archive.

        Returns
        -------
        Optional[Dict[str, Any]]
            The content of the CONFIG file -- None if there is no such file.
        """
        mfcontent_info = cls._extract_archive_mfcontent_info(archive, packaging_backend)
        handling_cls = cls._get_mfcontent_class(mfcontent_info)
        return handling_cls.get_archive_config_impl(
            mfcontent_info, archive, packaging_backend
        )

    @classmethod
    def get_archive_filename(
        cls,
        archive: Any,
        filename: str,
        content_type: ContentType,
        packaging_backend: Type[PackagingBackend] = TarPackagingBackend,
    ) -> Optional[str]:
        """
        Get the filename of the archive. This does not do any extraction but simply
        returns where, in the archive, the file is located. This is the equivalent of
        get_filename but for files not extracted yet.

        Parameters
        ----------
        archive: Any
            The archive to get the filename from.
        filename: str
            The name of the file in the archive.
        content_type: ContentType
            The type of the content (e.g., code, other, etc.).
        packaging_backend: Type[PackagingBackend], default TarPackagingBackend
            The packaging backend to use.

        Returns
        -------
        str
            The filename of the archive or None if not found.
        """
        mfcontent_info = cls._extract_archive_mfcontent_info(archive, packaging_backend)
        handling_cls = cls._get_mfcontent_class(mfcontent_info)
        return handling_cls.get_archive_filename_impl(
            mfcontent_info, archive, filename, content_type, packaging_backend
        )

    @classmethod
    def get_archive_content_names(
        cls,
        archive: Any,
        content_types: Optional[int] = None,
        packaging_backend: Type[PackagingBackend] = TarPackagingBackend,
    ) -> List[str]:
        mfcontent_info = cls._extract_archive_mfcontent_info(archive, packaging_backend)
        handling_cls = cls._get_mfcontent_class(mfcontent_info)
        return handling_cls.get_archive_content_names_impl(
            mfcontent_info, archive, content_types, packaging_backend
        )

    @classmethod
    def get_distribution_finder(
        cls,
    ) -> Optional["metaflow.extension_support.metadata.DistributionFinder"]:
        """
        Get the distribution finder for the Metaflow code package (if applicable).

        Some packages will include distribution information to "pretend" that some packages
        are actually distributions even if we just include them in the code package.

        Returns
        -------
        Optional["metaflow.extension_support.metadata.DistributionFinder"]
            The distribution finder for the Metaflow code package -- None if there is no
            such finder.
        """
        mfcontent_info = cls._extract_mfcontent_info()
        handling_cls = cls._get_mfcontent_class(mfcontent_info)
        return handling_cls.get_distribution_finder_impl(mfcontent_info)

    @classmethod
    def get_post_extract_env_vars(
        cls, version_id: int, dest_dir: str = "."
    ) -> Dict[str, str]:
        """
        Get the post-extract environment variables that are needed to access the content
        that has been extracted into dest_dir.

        This will typically involve setting PYTHONPATH.

        Parameters
        ----------
        version_id: int
            The version of MetaflowCodeContent for this package.
        dest_dir: str, default "."
            The directory where the content has been extracted to.

        Returns
        -------
        Dict[str, str]
            The post-extract environment variables that are needed to access the content
            that has been extracted into extracted_dir.
        """
        if version_id not in cls._mappings:
            raise ValueError(
                "Invalid package -- unknown version %s in info: %s"
                % (version_id, cls._mappings)
            )
        return cls._mappings[version_id].get_post_extract_env_vars_impl(dest_dir)

    # Implement the _impl methods in the base subclass (in this file). These need to
    # happen with as few imports as possible to prevent circular dependencies.
    @classmethod
    def get_info_impl(
        cls, mfcontent_info: Optional[Dict[str, Any]]
    ) -> Optional[Dict[str, Any]]:
        raise NotImplementedError("get_info_impl not implemented")

    @classmethod
    def get_config_impl(
        cls, mfcontent_info: Optional[Dict[str, Any]]
    ) -> Optional[Dict[str, Any]]:
        raise NotImplementedError("get_config_impl not implemented")

    @classmethod
    def get_filename_impl(
        cls,
        mfcontent_info: Optional[Dict[str, Any]],
        filename: str,
        content_type: ContentType,
    ) -> Optional[str]:
        raise NotImplementedError("get_filename_impl not implemented")

    @classmethod
    def get_distribution_finder_impl(
        cls, mfcontent_info: Optional[Dict[str, Any]]
    ) -> Optional["metaflow.extension_support.metadata.DistributionFinder"]:
        raise NotImplementedError("get_distribution_finder_impl not implemented")

    @classmethod
    def get_archive_info_impl(
        cls,
        mfcontent_info: Optional[Dict[str, Any]],
        archive: Any,
        packaging_backend: Type[PackagingBackend] = TarPackagingBackend,
    ) -> Optional[Dict[str, Any]]:
        raise NotImplementedError("get_archive_info_impl not implemented")

    @classmethod
    def get_archive_config_impl(
        cls,
        mfcontent_info: Optional[Dict[str, Any]],
        archive: Any,
        packaging_backend: Type[PackagingBackend] = TarPackagingBackend,
    ) -> Optional[Dict[str, Any]]:
        raise NotImplementedError("get_archive_config_impl not implemented")

    @classmethod
    def get_archive_filename_impl(
        cls,
        mfcontent_info: Optional[Dict[str, Any]],
        archive: Any,
        filename: str,
        content_type: ContentType,
        packaging_backend: Type[PackagingBackend] = TarPackagingBackend,
    ) -> Optional[str]:
        raise NotImplementedError("get_archive_filename_impl not implemented")

    @classmethod
    def get_archive_content_names_impl(
        cls,
        mfcontent_info: Optional[Dict[str, Any]],
        archive: Any,
        content_types: Optional[int] = None,
        packaging_backend: Type[PackagingBackend] = TarPackagingBackend,
    ) -> List[str]:
        raise NotImplementedError("get_archive_content_names_impl not implemented")

    @classmethod
    def get_post_extract_env_vars_impl(cls, dest_dir: str) -> Dict[str, str]:
        raise NotImplementedError("get_post_extract_env_vars_impl not implemented")

    def __init_subclass__(cls, version_id, **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        if version_id in MetaflowCodeContent._mappings:
            raise ValueError(
                "Version ID %s already exists in MetaflowCodeContent mappings "
                "-- this is a bug in Metaflow." % str(version_id)
            )
        MetaflowCodeContent._mappings[version_id] = cls
        cls._version_id = version_id

    # Implement these methods in sub-classes of the base sub-classes. These methods
    # are called later and can have more dependencies and so can live in other files.
    def get_excluded_tl_entries(self) -> List[str]:
        """
        When packaging Metaflow from within an executing Metaflow flow, we need to
        exclude the files that are inserted by this content from being packaged (possibly).

        Use this function to return these files or top-level directories.

        Returns
        -------
        List[str]
            Files or directories to exclude
        """
        return []

    def content_names(
        self, content_types: Optional[int] = None
    ) -> Generator[Tuple[str, str], None, None]:
        """
        Detailed list of the content of this MetaflowCodeContent. This will list all files
        (or non files -- for the INFO or CONFIG data for example) present in the archive.

        Parameters
        ----------
        content_types : Optional[int]
            The type of content to get the names of. If None, all content is returned.

        Yields
        ------
        Generator[Tuple[str, str], None, None]
            Path on the filesystem and the name in the archive
        """
        raise NotImplementedError("content_names not implemented")

    def contents(
        self, content_types: Optional[int] = None
    ) -> Generator[Tuple[Union[bytes, str], str], None, None]:
        """
        Very similar to content_names but returns the content of the non-files
        as well as bytes. For files, identical output as content_names

        Parameters
        ----------
        content_types : Optional[int]
            The type of content to get the content of. If None, all content is returned.

        Yields
        ------
        Generator[Tuple[Union[str, bytes], str], None, None]
            Content of the MF content
        """
        raise NotImplementedError("content not implemented")

    def show(self) -> str:
        """
        Returns a more human-readable string representation of the content of this
        MetaflowCodeContent. This will not, for example, list all files but summarize what
        is included at a more high level.

        Returns
        -------
        str
            A human-readable string representation of the content of this MetaflowCodeContent
        """
        raise NotImplementedError("show not implemented")

    def add_info(self, info: Dict[str, Any]) -> None:
        """
        Add the content of the INFO file to the Metaflow content

        Parameters
        ----------
        info: Dict[str, Any]
            The content of the INFO file
        """
        raise NotImplementedError("add_info not implemented")

    def add_config(self, config: Dict[str, Any]) -> None:
        """
        Add the content of the CONFIG file to the Metaflow content

        Parameters
        ----------
        config: Dict[str, Any]
            The content of the CONFIG file
        """
        raise NotImplementedError("add_config not implemented")

    def add_module(self, module_path: ModuleType) -> None:
        """
        Add a python module to the Metaflow content

        Parameters
        ----------
        module_path: ModuleType
            The module to add
        """
        raise NotImplementedError("add_module not implemented")

    def add_code_file(self, file_path: str, file_name: str) -> None:
        """
        Add a code file to the Metaflow content

        Parameters
        ----------
        file_path: str
            The path to the code file to add (on the filesystem)
        file_name: str
            The path in the archive to add the code file to
        """
        raise NotImplementedError("add_code_file not implemented")

    def add_other_file(self, file_path: str, file_name: str) -> None:
        """
        Add a non-python file to the Metaflow content

        Parameters
        ----------
        file_path: str
            The path to the file to add (on the filesystem)
        file_name: str
            The path in the archive to add the file to
        """
        raise NotImplementedError("add_other_file not implemented")

    @classmethod
    def _get_mfcontent_class(
        cls, info: Optional[Dict[str, Any]]
    ) -> Type["MetaflowCodeContent"]:
        if info is None:
            return MetaflowCodeContentV0
        if "version" not in info:
            raise ValueError("Invalid package -- missing version in info: %s" % info)
        version = info["version"]
        if version not in cls._mappings:
            raise ValueError(
                "Invalid package -- unknown version %s in info: %s" % (version, info)
            )

        return cls._mappings[version]

    @classmethod
    def _extract_archive_mfcontent_info(
        cls,
        archive: Any,
        packaging_backend: Type[PackagingBackend] = TarPackagingBackend,
    ) -> Optional[Dict[str, Any]]:
        if id(archive) in cls._cached_mfcontent_info:
            return cls._cached_mfcontent_info[id(archive)]

        mfcontent_info = None  # type: Optional[Dict[str, Any]]
        # Here we need to extract the information from the archive
        if packaging_backend.cls_has_member(archive, MFCONTENT_MARKER):
            # The MFCONTENT_MARKER file is present in the archive
            # We can extract the information from it
            extracted_info = packaging_backend.cls_get_member(archive, MFCONTENT_MARKER)
            if extracted_info:
                mfcontent_info = json.loads(extracted_info)
        cls._cached_mfcontent_info[id(archive)] = mfcontent_info
        return mfcontent_info

    @classmethod
    def _extract_mfcontent_info(cls) -> Optional[Dict[str, Any]]:
        if "_local" in cls._cached_mfcontent_info:
            return cls._cached_mfcontent_info["_local"]

        mfcontent_info = None  # type: Optional[Dict[str, Any]]
        if os.path.exists(os.path.join(get_metaflow_root(), MFCONTENT_MARKER)):
            with open(
                os.path.join(get_metaflow_root(), MFCONTENT_MARKER),
                "r",
                encoding="utf-8",
            ) as f:
                mfcontent_info = json.load(f)
        cls._cached_mfcontent_info["_local"] = mfcontent_info
        return mfcontent_info

    def get_package_version(self) -> int:
        """
        Get the version of MetaflowCodeContent for this package.
        """
        # _version_id is set in __init_subclass__ when the subclass is created
        return self._version_id


class MetaflowCodeContentV0(MetaflowCodeContent, version_id=0):
    @classmethod
    def get_info_impl(
        cls, mfcontent_info: Optional[Dict[str, Any]]
    ) -> Optional[Dict[str, Any]]:
        path_to_file = os.path.join(get_metaflow_root(), "INFO")
        if os.path.isfile(path_to_file):
            with open(path_to_file, "r", encoding="utf-8") as f:
                return json.load(f)
        return None

    @classmethod
    def get_config_impl(
        cls, mfcontent_info: Optional[Dict[str, Any]]
    ) -> Optional[Dict[str, Any]]:
        path_to_file = os.path.join(get_metaflow_root(), "CONFIG")
        if os.path.isfile(path_to_file):
            with open(path_to_file, "r", encoding="utf-8") as f:
                return json.load(f)
        return None

    @classmethod
    def get_filename_impl(
        cls,
        mfcontent_info: Optional[Dict[str, Any]],
        filename: str,
        content_type: ContentType,
    ) -> Optional[str]:
        """
        For V0, the filename is simply the filename passed in.
        """
        path_to_file = os.path.join(get_metaflow_root(), filename)
        if os.path.isfile(path_to_file):
            return path_to_file
        return None

    @classmethod
    def get_distribution_finder_impl(
        cls, mfcontent_info: Optional[Dict[str, Any]]
    ) -> Optional["metaflow.extension_support.metadata.DistributionFinder"]:
        return None

    @classmethod
    def get_archive_info_impl(
        cls,
        mfcontent_info: Optional[Dict[str, Any]],
        archive: Any,
        packaging_backend: Type[PackagingBackend] = TarPackagingBackend,
    ) -> Optional[Dict[str, Any]]:
        info_content = packaging_backend.cls_get_member(archive, "INFO")
        if info_content:
            return json.loads(info_content)
        return None

    @classmethod
    def get_archive_config_impl(
        cls,
        mfcontent_info: Optional[Dict[str, Any]],
        archive: Any,
        packaging_backend: Type[PackagingBackend] = TarPackagingBackend,
    ) -> Optional[Dict[str, Any]]:
        info_content = packaging_backend.cls_get_member(archive, "CONFIG")
        if info_content:
            return json.loads(info_content)
        return None

    @classmethod
    def get_archive_filename_impl(
        cls,
        mfcontent_info: Optional[Dict[str, Any]],
        archive: Any,
        filename: str,
        content_type: ContentType,
        packaging_backend: Type[PackagingBackend] = TarPackagingBackend,
    ) -> str:
        if packaging_backend.cls_has_member(archive, filename):
            # The file is present in the archive
            return filename
        return None

    @classmethod
    def get_archive_content_names_impl(
        cls,
        mfcontent_info: Optional[Dict[str, Any]],
        archive: Any,
        content_types: Optional[int] = None,
        packaging_backend: Type[PackagingBackend] = TarPackagingBackend,
    ) -> List[str]:
        """
        For V0, we use a static list of known files to classify the content
        """
        known_prefixes = {
            "metaflow/": ContentType.CODE_CONTENT.value,
            "metaflow_extensions/": ContentType.CODE_CONTENT.value,
            "INFO": ContentType.OTHER_CONTENT.value,
            "CONFIG": ContentType.OTHER_CONTENT.value,
            "conda.manifest": ContentType.OTHER_CONTENT.value,
            "uv.lock": ContentType.OTHER_CONTENT.value,
            "pyproject.toml": ContentType.OTHER_CONTENT.value,
            # Used in nflx-metaflow-extensions
            "condav2-1.cnd": ContentType.OTHER_CONTENT.value,
        }
        to_return = []
        for filename in packaging_backend.cls_list_members(archive):
            for prefix, classification in known_prefixes.items():
                if (
                    prefix[-1] == "/" and filename.startswith(prefix)
                ) or prefix == filename:
                    if content_types & classification:
                        to_return.append(filename)
                elif content_types & ContentType.USER_CONTENT.value:
                    # Everything else is user content
                    to_return.append(filename)
        return to_return

    @classmethod
    def get_post_extract_env_vars_impl(cls, dest_dir: str) -> Dict[str, str]:
        return {"PYTHONPATH": dest_dir}

    def get_excluded_tl_entries(self) -> List[str]:
        """
        When packaging Metaflow from within an executing Metaflow flow, we need to
        exclude the files that are inserted by this content from being packaged (possibly).

        Use this function to return these files or top-level directories.

        Returns
        -------
        List[str]
            Files or directories to exclude
        """
        return ["CONFIG", "INFO"]

    # Other non-implemented methods are OK not being implemented as they will never
    # be called as they are only used when creating the package and we are starting
    # with V1.


class MetaflowCodeContentV1Base(MetaflowCodeContent, version_id=1):
    _code_dir = ".mf_code"
    _other_dir = ".mf_meta"
    _info_file = "INFO"
    _config_file = "CONFIG"
    _dist_info_file = "DIST_INFO"

    def __init_subclass__(cls, **kwargs) -> None:
        # Important to add this here to prevent the subclass of MetaflowCodeContentV1Base from
        # also calling __init_subclass__ in MetaflowCodeContent (which would create a problem)
        return None

    def __init__(self, code_dir: str, other_dir: str) -> None:
        self._code_dir = code_dir
        self._other_dir = other_dir

    @classmethod
    def _get_otherfile_path(
        cls, mfcontent_info: Optional[Dict[str, Any]], filename: str, in_archive: bool
    ) -> str:
        if in_archive:
            return filename
        return os.path.join(get_metaflow_root(), "..", cls._other_dir, filename)

    @classmethod
    def _get_codefile_path(
        cls, mfcontent_info: Optional[Dict[str, Any]], filename: str, in_archive: bool
    ) -> str:
        if in_archive:
            return filename
        return os.path.join(get_metaflow_root(), filename)

    @classmethod
    def get_info_impl(
        cls, mfcontent_info: Optional[Dict[str, Any]]
    ) -> Optional[Dict[str, Any]]:
        path_to_file = cls._get_otherfile_path(
            mfcontent_info, cls._info_file, in_archive=False
        )
        if os.path.isfile(path_to_file):
            with open(path_to_file, "r", encoding="utf-8") as f:
                return json.load(f)
        return None

    @classmethod
    def get_config_impl(
        cls, mfcontent_info: Optional[Dict[str, Any]]
    ) -> Optional[Dict[str, Any]]:
        path_to_file = cls._get_otherfile_path(
            mfcontent_info, cls._config_file, in_archive=False
        )
        if os.path.isfile(path_to_file):
            with open(path_to_file, "r", encoding="utf-8") as f:
                return json.load(f)
        return None

    @classmethod
    def get_filename_impl(
        cls,
        mfcontent_info: Optional[Dict[str, Any]],
        filename: str,
        content_type: ContentType,
    ) -> Optional[str]:
        if content_type == ContentType.CODE_CONTENT:
            path_to_file = cls._get_codefile_path(
                mfcontent_info, filename, in_archive=False
            )
        elif content_type in (ContentType.OTHER_CONTENT, ContentType.MODULE_CONTENT):
            path_to_file = cls._get_otherfile_path(
                mfcontent_info, filename, in_archive=False
            )
        else:
            raise ValueError(
                f"Invalid content type {content_type} for filename {filename}"
            )
        if os.path.isfile(path_to_file):
            return path_to_file
        return None

    @classmethod
    def get_distribution_finder_impl(
        cls, mfcontent_info: Optional[Dict[str, Any]]
    ) -> Optional["metaflow.extension_support.metadata.DistributionFinder"]:
        path_to_file = cls._get_otherfile_path(
            mfcontent_info, cls._dist_info_file, in_archive=False
        )
        if os.path.isfile(path_to_file):
            with open(path_to_file, "r", encoding="utf-8") as f:
                return PackagedDistributionFinder(json.load(f))
        return None

    @classmethod
    def get_archive_info_impl(
        cls,
        mfcontent_info: Optional[Dict[str, Any]],
        archive: Any,
        packaging_backend: Type[PackagingBackend] = TarPackagingBackend,
    ) -> Optional[Dict[str, Any]]:
        info_file = packaging_backend.cls_get_member(
            archive,
            cls._get_otherfile_path(mfcontent_info, cls._info_file, in_archive=True),
        )
        if info_file:
            return json.loads(info_file)
        return None

    @classmethod
    def get_archive_config_impl(
        cls,
        mfcontent_info: Optional[Dict[str, Any]],
        archive: Any,
        packaging_backend: Type[PackagingBackend] = TarPackagingBackend,
    ) -> Optional[Dict[str, Any]]:
        config_file = packaging_backend.cls_get_member(
            archive,
            cls._get_otherfile_path(mfcontent_info, cls._config_file, in_archive=True),
        )
        if config_file:
            return json.loads(config_file)
        return None

    @classmethod
    def get_archive_filename_impl(
        cls,
        mfcontent_info: Optional[Dict[str, Any]],
        archive: Any,
        filename: str,
        content_type: ContentType,
        packaging_backend: Type[PackagingBackend] = TarPackagingBackend,
    ) -> str:
        if content_type == ContentType.CODE_CONTENT:
            path_to_file = cls._get_codefile_path(
                mfcontent_info, filename, in_archive=False
            )
        elif content_type in (ContentType.OTHER_CONTENT, ContentType.MODULE_CONTENT):
            path_to_file = cls._get_otherfile_path(
                mfcontent_info, filename, in_archive=False
            )
        else:
            raise ValueError(
                f"Invalid content type {content_type} for filename {filename}"
            )
        if packaging_backend.cls_has_member(archive, path_to_file):
            # The file is present in the archive
            return path_to_file
        return None

    @classmethod
    def get_archive_content_names_impl(
        cls,
        mfcontent_info: Optional[Dict[str, Any]],
        archive: Any,
        content_types: Optional[int] = None,
        packaging_backend: Type[PackagingBackend] = TarPackagingBackend,
    ) -> List[str]:
        to_return = []
        module_content = set(mfcontent_info.get("module_files", []))
        for filename in packaging_backend.cls_list_members(archive):
            if filename.startswith(cls._other_dir) and (
                content_types & ContentType.OTHER_CONTENT.value
            ):
                to_return.append(filename)
            elif filename.startswith(cls._code_dir):
                # Special case for marker which is a other content even if in code.
                if filename == f"{cls._code_dir}/{MFCONTENT_MARKER}":
                    if content_types & ContentType.OTHER_CONTENT.value:
                        to_return.append(filename)
                    else:
                        continue
                # Here it is either module or code
                if os.path.join(cls._code_dir, filename) in module_content:
                    if content_types & ContentType.MODULE_CONTENT.value:
                        to_return.append(filename)
                elif content_types & ContentType.CODE_CONTENT.value:
                    to_return.append(filename)
            else:
                if content_types & ContentType.USER_CONTENT.value:
                    # Everything else is user content
                    to_return.append(filename)
        return to_return

    @classmethod
    def get_post_extract_env_vars_impl(cls, dest_dir: str) -> Dict[str, str]:
        return {"PYTHONPATH": f"{dest_dir}/{cls._code_dir}"}



import base64
import json
import platform
import re
import sys
from hashlib import sha1
from time import sleep

from metaflow import JSONType, Run, current, decorators, parameters
from metaflow._vendor import click
from metaflow.exception import (
    MetaflowException,
    MetaflowInternalError,
    MetaflowNotFound,
)
from metaflow.metaflow_config import (
    ARGO_WORKFLOWS_UI_URL,
    FEAT_ALWAYS_UPLOAD_CODE_PACKAGE,
    KUBERNETES_NAMESPACE,
    SERVICE_VERSION_CHECK,
    UI_URL,
)
from metaflow.package import MetaflowPackage

# TODO: Move production_token to utils
from metaflow.plugins.aws.step_functions.production_token import (
    load_token,
    new_token,
    store_token,
)
from metaflow.plugins.environment_decorator import EnvironmentDecorator
from metaflow.plugins.kubernetes.kubernetes_decorator import KubernetesDecorator
from metaflow.tagging_util import validate_tags
from metaflow.util import get_username, to_bytes, to_unicode, version_parse

from .argo_workflows import ArgoWorkflows

VALID_NAME = re.compile(r"^[a-z0-9]([a-z0-9\.\-]*[a-z0-9])?$")

unsupported_decorators = {
    "snowpark": "Step *%s* is marked for execution on Snowpark with Argo Workflows which isn't currently supported.",
    "slurm": "Step *%s* is marked for execution on Slurm with Argo Workflows which isn't currently supported.",
    "nvidia": "Step *%s* is marked for execution on Nvidia with Argo Workflows which isn't currently supported.",
    "nvct": "Step *%s* is marked for execution on Nvct with Argo Workflows which isn't currently supported.",
}


class IncorrectProductionToken(MetaflowException):
    headline = "Incorrect production token"


class RunIdMismatch(MetaflowException):
    headline = "Run ID mismatch"


class IncorrectMetadataServiceVersion(MetaflowException):
    headline = "Incorrect version for metaflow service"


class ArgoWorkflowsNameTooLong(MetaflowException):
    headline = "Argo Workflows name too long"


class UnsupportedPythonVersion(MetaflowException):
    headline = "Unsupported version of Python"


@click.group()
def cli():
    pass


@cli.group(help="Commands related to Argo Workflows.")
@click.option(
    "--name",
    default=None,
    type=str,
    help="Argo Workflow name. The flow name is used instead if "
    "this option is not specified.",
)
@click.pass_obj
def argo_workflows(obj, name=None):
    check_python_version(obj)
    obj.check(obj.graph, obj.flow, obj.environment, pylint=obj.pylint)
    (
        obj.workflow_name,
        obj.token_prefix,
        obj.is_project,
    ) = resolve_workflow_name(obj, name)


@argo_workflows.command(help="Deploy a new version of this workflow to Argo Workflows.")
@click.option(
    "--authorize",
    default=None,
    help="Authorize using this production token. You need this "
    "when you are re-deploying an existing flow for the first "
    "time. The token is cached in METAFLOW_HOME, so you only "
    "need to specify this once.",
)
@click.option(
    "--generate-new-token",
    is_flag=True,
    help="Generate a new production token for this flow. "
    "This will move the production flow to a new namespace.",
)
@click.option(
    "--new-token",
    "given_token",
    default=None,
    help="Use the given production token for this flow. "
    "This will move the production flow to the given namespace.",
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    default=None,
    help="Annotate all objects produced by Argo Workflows runs "
    "with the given tag. You can specify this option multiple "
    "times to attach multiple tags.",
)
@click.option(
    "--namespace",
    "user_namespace",
    default=None,
    help="Change the namespace from the default (production token) "
    "to the given tag. See run --help for more information.",
)
@click.option(
    "--only-json",
    is_flag=True,
    default=False,
    help="Only print out JSON sent to Argo Workflows. Do not deploy anything.",
    hidden=True,
)
@click.option(
    "--max-workers",
    default=100,
    show_default=True,
    help="Maximum number of parallel processes.",
)
@click.option(
    "--workflow-timeout", default=None, type=int, help="Workflow timeout in seconds."
)
@click.option(
    "--workflow-priority",
    default=None,
    type=int,
    help="Workflow priority as an integer. Workflows with higher priority "
    "are processed first if Argo Workflows controller is configured to process "
    "limited number of workflows in parallel",
)
@click.option(
    "--auto-emit-argo-events/--no-auto-emit-argo-events",
    default=True,  # TODO: Default to a value from config
    show_default=True,
    help="Auto emits Argo Events when the run completes successfully.",
)
@click.option(
    "--notify-on-error/--no-notify-on-error",
    default=False,
    show_default=True,
    help="Notify if the workflow fails.",
)
@click.option(
    "--notify-on-success/--no-notify-on-success",
    default=False,
    show_default=True,
    help="Notify if the workflow succeeds.",
)
@click.option(
    "--notify-slack-webhook-url",
    default=None,
    help="Slack incoming webhook url for workflow success/failure notifications.",
)
@click.option(
    "--notify-pager-duty-integration-key",
    default=None,
    help="PagerDuty Events API V2 Integration key for workflow success/failure notifications.",
)
@click.option(
    "--notify-incident-io-api-key",
    default=None,
    help="Incident.io API V2 key for workflow success/failure notifications.",
)
@click.option(
    "--incident-io-alert-source-config-id",
    default=None,
    help="Incident.io Alert source config ID. Example '01GW2G3V0S59R238FAHPDS1R66'",
)
@click.option(
    "--incident-io-metadata",
    default=None,
    type=str,
    multiple=True,
    help="Incident.io Alert Custom Metadata field in the form of Key=Value",
)
@click.option(
    "--enable-heartbeat-daemon/--no-enable-heartbeat-daemon",
    default=False,
    show_default=True,
    help="Use a daemon container to broadcast heartbeats.",
)
@click.option(
    "--deployer-attribute-file",
    default=None,
    show_default=True,
    type=str,
    help="Write the workflow name to the file specified. Used internally for Metaflow's Deployer API.",
    hidden=True,
)
@click.option(
    "--enable-error-msg-capture/--no-enable-error-msg-capture",
    default=True,
    show_default=True,
    help="Capture stack trace of first failed task in exit hook.",
)
@click.pass_obj
def create(
    obj,
    tags=None,
    user_namespace=None,
    only_json=False,
    authorize=None,
    generate_new_token=False,
    given_token=None,
    max_workers=None,
    workflow_timeout=None,
    workflow_priority=None,
    auto_emit_argo_events=False,
    notify_on_error=False,
    notify_on_success=False,
    notify_slack_webhook_url=None,
    notify_pager_duty_integration_key=None,
    notify_incident_io_api_key=None,
    incident_io_alert_source_config_id=None,
    incident_io_metadata=None,
    enable_heartbeat_daemon=True,
    deployer_attribute_file=None,
    enable_error_msg_capture=False,
):
    for node in obj.graph:
        for decorator, error_message in unsupported_decorators.items():
            if any([d.name == decorator for d in node.decorators]):
                raise MetaflowException(error_message % node.name)

    validate_tags(tags)

    if deployer_attribute_file:
        with open(deployer_attribute_file, "w", encoding="utf-8") as f:
            json.dump(
                {
                    "name": obj.workflow_name,
                    "flow_name": obj.flow.name,
                    "metadata": obj.metadata.metadata_str(),
                },
                f,
            )

    obj.echo("Deploying *%s* to Argo Workflows..." % obj.workflow_name, bold=True)

    if SERVICE_VERSION_CHECK:
        # TODO: Consider dispelling with this check since it's been 2 years since the
        #       needed metadata service changes have been available in open-source. It's
        #       likely that Metaflow users may not have access to metadata service from
        #       within their workstations.
        check_metadata_service_version(obj)

    token = resolve_token(
        obj.workflow_name,
        obj.token_prefix,
        obj,
        authorize,
        given_token,
        generate_new_token,
        obj.is_project,
    )

    flow = make_flow(
        obj,
        token,
        obj.workflow_name,
        tags,
        user_namespace,
        max_workers,
        workflow_timeout,
        workflow_priority,
        auto_emit_argo_events,
        notify_on_error,
        notify_on_success,
        notify_slack_webhook_url,
        notify_pager_duty_integration_key,
        notify_incident_io_api_key,
        incident_io_alert_source_config_id,
        incident_io_metadata,
        enable_heartbeat_daemon,
        enable_error_msg_capture,
    )

    if only_json:
        obj.echo_always(str(flow), err=False, no_bold=True)
        # TODO: Support echo-ing Argo Events Sensor template
    else:
        flow.deploy()
        obj.echo(
            "Workflow *{workflow_name}* "
            "for flow *{name}* pushed to "
            "Argo Workflows successfully.\n".format(
                workflow_name=obj.workflow_name, name=current.flow_name
            ),
            bold=True,
        )
        if obj._is_workflow_name_modified:
            obj.echo(
                "Note that the flow was deployed with a modified name "
                "due to Kubernetes naming conventions\non Argo Workflows. The "
                "original flow name is stored in the workflow annotation.\n"
            )

        if ARGO_WORKFLOWS_UI_URL:
            obj.echo("See the deployed workflow here:", bold=True)
            argo_workflowtemplate_link = "%s/workflow-templates/%s" % (
                ARGO_WORKFLOWS_UI_URL.rstrip("/"),
                KUBERNETES_NAMESPACE,
            )
            obj.echo(
                "%s/%s\n\n" % (argo_workflowtemplate_link, obj.workflow_name),
                indent=True,
            )
        flow.schedule()
        obj.echo("What will trigger execution of the workflow:", bold=True)
        obj.echo(flow.trigger_explanation(), indent=True)

        # TODO: Print events emitted by execution of this flow

        # response = ArgoWorkflows.trigger(obj.workflow_name)
        # run_id = "argo-" + response["metadata"]["name"]

        # obj.echo(
        #     "Workflow *{name}* triggered on Argo Workflows "
        #     "(run-id *{run_id}*).".format(name=obj.workflow_name, run_id=run_id),
        #     bold=True,
        # )


def check_python_version(obj):
    # argo-workflows integration for Metaflow isn't supported for Py versions below 3.6.
    # This constraint can very well be lifted if desired.
    if sys.version_info < (3, 6):
        obj.echo("")
        obj.echo(
            "Metaflow doesn't support Argo Workflows for Python %s right now."
            % platform.python_version()
        )
        obj.echo(
            "Please upgrade your Python interpreter to version 3.6 (or higher) or "
            "reach out to us at slack.outerbounds.co for more help."
        )
        raise UnsupportedPythonVersion(
            "Try again with a more recent version of Python (>=3.6)."
        )


def check_metadata_service_version(obj):
    metadata = obj.metadata
    version = metadata.version()
    if version == "local":
        return
    elif version is not None and version_parse(version) >= version_parse("2.0.2"):
        # Metaflow metadata service needs to be at least at version 2.0.2
        # since prior versions did not support strings as object ids.
        return
    else:
        obj.echo("")
        obj.echo(
            "You are running a version of the metaflow service that currently doesn't "
            "support Argo Workflows. "
        )
        obj.echo(
            "For more information on how to upgrade your service to a compatible "
            "version (>= 2.0.2), visit:"
        )
        obj.echo(
            "    https://admin-docs.metaflow.org/metaflow-on-aws/operation"
            "s-guide/metaflow-service-migration-guide",
            fg="green",
        )
        obj.echo(
            "Once you have upgraded your metadata service, please re-execute your "
            "command."
        )
        raise IncorrectMetadataServiceVersion(
            "Try again with a more recent version of metaflow service (>=2.0.2)."
        )


def resolve_workflow_name(obj, name):
    project = current.get("project_name")
    obj._is_workflow_name_modified = False
    if project:
        if name:
            raise MetaflowException(
                "--name is not supported for @projects. Use --branch instead."
            )
        workflow_name = current.project_flow_name
        project_branch = to_bytes(".".join((project, current.branch_name)))
        token_prefix = (
            "mfprj-%s"
            % to_unicode(base64.b32encode(sha1(project_branch).digest()))[:16]
        )
        is_project = True
        # Argo Workflow names can't be longer than 253 characters, so we truncate
        # by default. Also, while project and branch allow for underscores, Argo
        # Workflows doesn't (DNS Subdomain names as defined in RFC 1123) - so we will
        # remove any underscores as well as convert the name to lower case.
        # Also remove + and @ as not allowed characters, which can be part of the
        # project branch due to using email addresses as user names.
        if len(workflow_name) > 253:
            name_hash = to_unicode(
                base64.b32encode(sha1(to_bytes(workflow_name)).digest())
            )[:8].lower()
            workflow_name = "%s-%s" % (workflow_name[:242], name_hash)
            obj._is_workflow_name_modified = True
        if not VALID_NAME.search(workflow_name):
            workflow_name = sanitize_for_argo(workflow_name)
            obj._is_workflow_name_modified = True
    else:
        if name and not VALID_NAME.search(name):
            raise MetaflowException(
                "Name '%s' contains invalid characters. The "
                "name must consist of lower case alphanumeric characters, '-' or '.'"
                ", and must start and end with an alphanumeric character." % name
            )

        workflow_name = name if name else current.flow_name
        token_prefix = workflow_name
        is_project = False

        if len(workflow_name) > 253:
            msg = (
                "The full name of the workflow:\n*%s*\nis longer than 253 "
                "characters.\n\n"
                "To deploy this workflow to Argo Workflows, please "
                "assign a shorter name\nusing the option\n"
                "*argo-workflows --name <name> create*." % workflow_name
            )
            raise ArgoWorkflowsNameTooLong(msg)

        if not VALID_NAME.search(workflow_name):
            workflow_name = sanitize_for_argo(workflow_name)
            obj._is_workflow_name_modified = True

    return workflow_name, token_prefix.lower(), is_project


def make_flow(
    obj,
    token,
    name,
    tags,
    namespace,
    max_workers,
    workflow_timeout,
    workflow_priority,
    auto_emit_argo_events,
    notify_on_error,
    notify_on_success,
    notify_slack_webhook_url,
    notify_pager_duty_integration_key,
    notify_incident_io_api_key,
    incident_io_alert_source_config_id,
    incident_io_metadata,
    enable_heartbeat_daemon,
    enable_error_msg_capture,
):
    # TODO: Make this check less specific to Amazon S3 as we introduce
    #       support for more cloud object stores.
    if obj.flow_datastore.TYPE not in ("azure", "gs", "s3"):
        raise MetaflowException(
            "Argo Workflows requires --datastore=s3 or --datastore=azure or --datastore=gs"
        )

    if (notify_on_error or notify_on_success) and not (
        notify_slack_webhook_url
        or notify_pager_duty_integration_key
        or notify_incident_io_api_key
    ):
        raise MetaflowException(
            "Notifications require specifying an incoming Slack webhook url via --notify-slack-webhook-url, PagerDuty events v2 integration key via --notify-pager-duty-integration-key or\n"
            "Incident.io integration API key via --notify-incident-io-api-key.\n"
            " If you would like to set up notifications for your Slack workspace, follow the instructions at "
            "https://api.slack.com/messaging/webhooks to generate a webhook url.\n"
            " For notifications through PagerDuty, generate an integration key by following the instructions at "
            "https://support.pagerduty.com/docs/services-and-integrations#create-a-generic-events-api-integration\n"
            " For notifications through Incident.io, generate an alert source config."
        )

    if (
        (notify_on_error or notify_on_success)
        and notify_incident_io_api_key
        and incident_io_alert_source_config_id is None
    ):
        raise MetaflowException(
            "Incident.io alerts require an alert source configuration ID. Please set one with --incident-io-alert-source-config-id"
        )

    # Attach @kubernetes and @environment decorator to the flow to
    # ensure that the related decorator hooks are invoked.
    decorators._attach_decorators(
        obj.flow, [KubernetesDecorator.name, EnvironmentDecorator.name]
    )
    decorators._init(obj.flow)

    decorators._init_step_decorators(
        obj.flow, obj.graph, obj.environment, obj.flow_datastore, obj.logger
    )
    obj.graph = obj.flow._graph

    # Save the code package in the flow datastore so that both user code and
    # metaflow package can be retrieved during workflow execution.
    obj.package = MetaflowPackage(
        obj.flow,
        obj.environment,
        obj.echo,
        suffixes=obj.package_suffixes,
        flow_datastore=obj.flow_datastore if FEAT_ALWAYS_UPLOAD_CODE_PACKAGE else None,
    )

    # This blocks until the package is created
    if FEAT_ALWAYS_UPLOAD_CODE_PACKAGE:
        package_url = obj.package.package_url()
        package_sha = obj.package.package_sha()
    else:
        package_url, package_sha = obj.flow_datastore.save_data(
            [obj.package.blob], len_hint=1
        )[0]

    return ArgoWorkflows(
        name,
        obj.graph,
        obj.flow,
        obj.package.package_metadata,
        package_sha,
        package_url,
        token,
        obj.metadata,
        obj.flow_datastore,
        obj.environment,
        obj.event_logger,
        obj.monitor,
        tags=tags,
        namespace=namespace,
        max_workers=max_workers,
        username=get_username(),
        workflow_timeout=workflow_timeout,
        workflow_priority=workflow_priority,
        auto_emit_argo_events=auto_emit_argo_events,
        notify_on_error=notify_on_error,
        notify_on_success=notify_on_success,
        notify_slack_webhook_url=notify_slack_webhook_url,
        notify_pager_duty_integration_key=notify_pager_duty_integration_key,
        notify_incident_io_api_key=notify_incident_io_api_key,
        incident_io_alert_source_config_id=incident_io_alert_source_config_id,
        incident_io_metadata=incident_io_metadata,
        enable_heartbeat_daemon=enable_heartbeat_daemon,
        enable_error_msg_capture=enable_error_msg_capture,
    )


# TODO: Unify this method with the one in step_functions_cli.py
def resolve_token(
    name, token_prefix, obj, authorize, given_token, generate_new_token, is_project
):
    # 1) retrieve the previous deployment, if one exists
    workflow = ArgoWorkflows.get_existing_deployment(name)
    if workflow is None:
        obj.echo(
            "It seems this is the first time you are deploying *%s* to "
            "Argo Workflows." % name
        )
        prev_token = None
    else:
        prev_user, prev_token = workflow

    # 2) authorize this deployment
    if prev_token is not None:
        if authorize is None:
            authorize = load_token(token_prefix)
        elif authorize.startswith("production:"):
            authorize = authorize[11:]

        # we allow the user who deployed the previous version to re-deploy,
        # even if they don't have the token
        if prev_user != get_username() and authorize != prev_token:
            obj.echo(
                "There is an existing version of *%s* on Argo Workflows which was "
                "deployed by the user *%s*." % (name, prev_user)
            )
            obj.echo(
                "To deploy a new version of this flow, you need to use the same "
                "production token that they used. "
            )
            obj.echo(
                "Please reach out to them to get the token. Once you have it, call "
                "this command:"
            )
            obj.echo("    argo-workflows create --authorize MY_TOKEN", fg="green")
            obj.echo(
                'See "Organizing Results" at docs.metaflow.org for more information '
                "about production tokens."
            )
            raise IncorrectProductionToken(
                "Try again with the correct production token."
            )

    # 3) do we need a new token or should we use the existing token?
    if given_token:
        if is_project:
            # we rely on a known prefix for @project tokens, so we can't
            # allow the user to specify a custom token with an arbitrary prefix
            raise MetaflowException(
                "--new-token is not supported for @projects. Use --generate-new-token "
                "to create a new token."
            )
        if given_token.startswith("production:"):
            given_token = given_token[11:]
        token = given_token
        obj.echo("")
        obj.echo("Using the given token, *%s*." % token)
    elif prev_token is None or generate_new_token:
        token = new_token(token_prefix, prev_token)
        if token is None:
            if prev_token is None:
                raise MetaflowInternalError(
                    "We could not generate a new token. This is unexpected. "
                )
            else:
                raise MetaflowException(
                    "--generate-new-token option is not supported after using "
                    "--new-token. Use --new-token to make a new namespace."
                )
        obj.echo("")
        obj.echo("A new production token generated.")
    else:
        token = prev_token

    obj.echo("")
    obj.echo("The namespace of this production flow is")
    obj.echo("    production:%s" % token, fg="green")
    obj.echo(
        "To analyze results of this production flow add this line in your notebooks:"
    )
    obj.echo('    namespace("production:%s")' % token, fg="green")
    obj.echo(
        "If you want to authorize other people to deploy new versions of this flow to "
        "Argo Workflows, they need to call"
    )
    obj.echo("    argo-workflows create --authorize %s" % token, fg="green")
    obj.echo("when deploying this flow to Argo Workflows for the first time.")
    obj.echo(
        'See "Organizing Results" at https://docs.metaflow.org/ for more '
        "information about production tokens."
    )
    obj.echo("")
    store_token(token_prefix, token)
    return token


@parameters.add_custom_parameters(deploy_mode=False)
@argo_workflows.command(help="Trigger the workflow on Argo Workflows.")
@click.option(
    "--run-id-file",
    default=None,
    show_default=True,
    type=str,
    help="Write the ID of this run to the file specified.",
)
@click.option(
    "--deployer-attribute-file",
    default=None,
    show_default=True,
    type=str,
    help="Write the metadata and pathspec of this run to the file specified.\nUsed internally for Metaflow's Deployer API.",
    hidden=True,
)
@click.pass_obj
def trigger(obj, run_id_file=None, deployer_attribute_file=None, **kwargs):
    def _convert_value(param):
        # Swap `-` with `_` in parameter name to match click's behavior
        val = kwargs.get(param.name.replace("-", "_").lower())
        if param.kwargs.get("type") == JSONType:
            val = json.dumps(val)
        elif isinstance(val, parameters.DelayedEvaluationParameter):
            val = val(return_str=True)
        return val

    params = {
        param.name: _convert_value(param)
        for _, param in obj.flow._get_parameters()
        if kwargs.get(param.name.replace("-", "_").lower()) is not None
    }

    response = ArgoWorkflows.trigger(obj.workflow_name, params)
    run_id = "argo-" + response["metadata"]["name"]

    if run_id_file:
        with open(run_id_file, "w") as f:
            f.write(str(run_id))

    if deployer_attribute_file:
        with open(deployer_attribute_file, "w") as f:
            json.dump(
                {
                    "name": obj.workflow_name,
                    "metadata": obj.metadata.metadata_str(),
                    "pathspec": "/".join((obj.flow.name, run_id)),
                },
                f,
            )

    obj.echo(
        "Workflow *{name}* triggered on Argo Workflows "
        "(run-id *{run_id}*).".format(name=obj.workflow_name, run_id=run_id),
        bold=True,
    )

    run_url = (
        "%s/%s/%s" % (UI_URL.rstrip("/"), obj.flow.name, run_id) if UI_URL else None
    )

    if run_url:
        obj.echo(
            "See the run in the UI at %s" % run_url,
            bold=True,
        )


@argo_workflows.command(help="Delete the flow on Argo Workflows.")
@click.option(
    "--authorize",
    default=None,
    type=str,
    help="Authorize the deletion with a production token",
)
@click.pass_obj
def delete(obj, authorize=None):
    def _token_instructions(flow_name, prev_user):
        obj.echo(
            "There is an existing version of *%s* on Argo Workflows which was "
            "deployed by the user *%s*." % (flow_name, prev_user)
        )
        obj.echo(
            "To delete this flow, you need to use the same production token that they used."
        )
        obj.echo(
            "Please reach out to them to get the token. Once you have it, call "
            "this command:"
        )
        obj.echo("    argo-workflows delete --authorize MY_TOKEN", fg="green")
        obj.echo(
            'See "Organizing Results" at docs.metaflow.org for more information '
            "about production tokens."
        )

    validate_token(obj.workflow_name, obj.token_prefix, authorize, _token_instructions)
    obj.echo("Deleting workflow *{name}*...".format(name=obj.workflow_name), bold=True)

    schedule_deleted, sensor_deleted, workflow_deleted = ArgoWorkflows.delete(
        obj.workflow_name
    )

    if schedule_deleted:
        obj.echo(
            "Deleting cronworkflow *{name}*...".format(name=obj.workflow_name),
            bold=True,
        )

    if sensor_deleted:
        obj.echo(
            "Deleting sensor *{name}*...".format(name=obj.workflow_name),
            bold=True,
        )

    if workflow_deleted:
        obj.echo(
            "Deleting Kubernetes resources may take a while. "
            "Deploying the flow again to Argo Workflows while the delete is in-flight will fail."
        )
        obj.echo(
            "In-flight executions will not be affected. "
            "If necessary, terminate them manually."
        )


@argo_workflows.command(help="Suspend flow execution on Argo Workflows.")
@click.option(
    "--authorize",
    default=None,
    type=str,
    help="Authorize the suspension with a production token",
)
@click.argument("run-id", required=True, type=str)
@click.pass_obj
def suspend(obj, run_id, authorize=None):
    def _token_instructions(flow_name, prev_user):
        obj.echo(
            "There is an existing version of *%s* on Argo Workflows which was "
            "deployed by the user *%s*." % (flow_name, prev_user)
        )
        obj.echo(
            "To suspend this flow, you need to use the same production token that they used."
        )
        obj.echo(
            "Please reach out to them to get the token. Once you have it, call "
            "this command:"
        )
        obj.echo("    argo-workflows suspend RUN_ID --authorize MY_TOKEN", fg="green")
        obj.echo(
            'See "Organizing Results" at docs.metaflow.org for more information '
            "about production tokens."
        )

    validate_run_id(
        obj.workflow_name, obj.token_prefix, authorize, run_id, _token_instructions
    )

    # Trim prefix from run_id
    name = run_id[5:]

    workflow_suspended = ArgoWorkflows.suspend(name)

    if workflow_suspended:
        obj.echo("Suspended execution of *%s*" % run_id)


@argo_workflows.command(help="Unsuspend flow execution on Argo Workflows.")
@click.option(
    "--authorize",
    default=None,
    type=str,
    help="Authorize the unsuspend with a production token",
)
@click.argument("run-id", required=True, type=str)
@click.pass_obj
def unsuspend(obj, run_id, authorize=None):
    def _token_instructions(flow_name, prev_user):
        obj.echo(
            "There is an existing version of *%s* on Argo Workflows which was "
            "deployed by the user *%s*." % (flow_name, prev_user)
        )
        obj.echo(
            "To unsuspend this flow, you need to use the same production token that they used."
        )
        obj.echo(
            "Please reach out to them to get the token. Once you have it, call "
            "this command:"
        )
        obj.echo(
            "    argo-workflows unsuspend RUN_ID --authorize MY_TOKEN",
            fg="green",
        )
        obj.echo(
            'See "Organizing Results" at docs.metaflow.org for more information '
            "about production tokens."
        )

    validate_run_id(
        obj.workflow_name, obj.token_prefix, authorize, run_id, _token_instructions
    )

    # Trim prefix from run_id
    name = run_id[5:]

    workflow_suspended = ArgoWorkflows.unsuspend(name)

    if workflow_suspended:
        obj.echo("Unsuspended execution of *%s*" % run_id)


def validate_token(name, token_prefix, authorize, instructions_fn=None):
    """
    Validate that the production token matches that of the deployed flow.
    In case both the user and token do not match, raises an error.
    Optionally outputs instructions on token usage via the provided instruction_fn(flow_name, prev_user)
    """
    # TODO: Unify this with the existing resolve_token implementation.

    # 1) retrieve the previous deployment, if one exists
    workflow = ArgoWorkflows.get_existing_deployment(name)
    if workflow is None:
        prev_token = None
    else:
        prev_user, prev_token = workflow

    # 2) authorize this deployment
    if prev_token is not None:
        if authorize is None:
            authorize = load_token(token_prefix)
        elif authorize.startswith("production:"):
            authorize = authorize[11:]

        # we allow the user who deployed the previous version to re-deploy,
        # even if they don't have the token
        # NOTE: The username is visible in multiple sources, and can be set by the user.
        # Should we consider being stricter here?
        if prev_user != get_username() and authorize != prev_token:
            if instructions_fn:
                instructions_fn(flow_name=name, prev_user=prev_user)
            raise IncorrectProductionToken(
                "Try again with the correct production token."
            )

    # 3) all validations passed, store the previous token for future use
    token = prev_token

    store_token(token_prefix, token)
    return True


def get_run_object(pathspec: str):
    try:
        return Run(pathspec, _namespace_check=False)
    except MetaflowNotFound:
        return None


def get_status_considering_run_object(status, run_obj):
    remapped_status = remap_status(status)
    if remapped_status == "Running" and run_obj is None:
        return "Pending"
    return remapped_status


@argo_workflows.command(help="Fetch flow execution status on Argo Workflows.")
@click.argument("run-id", required=True, type=str)
@click.pass_obj
def status(obj, run_id):
    if not run_id.startswith("argo-"):
        raise RunIdMismatch(
            "Run IDs for flows executed through Argo Workflows begin with 'argo-'"
        )
    obj.echo(
        "Fetching status for run *{run_id}* for {flow_name} ...".format(
            run_id=run_id, flow_name=obj.flow.name
        ),
        bold=True,
    )
    # Trim prefix from run_id
    name = run_id[5:]
    status = ArgoWorkflows.get_workflow_status(obj.flow.name, name)
    run_obj = get_run_object("/".join((obj.flow.name, run_id)))
    if status is not None:
        status = get_status_considering_run_object(status, run_obj)
        obj.echo_always(status)


@argo_workflows.command(help="Terminate flow execution on Argo Workflows.")
@click.option(
    "--authorize",
    default=None,
    type=str,
    help="Authorize the termination with a production token",
)
@click.argument("run-id", required=True, type=str)
@click.pass_obj
def terminate(obj, run_id, authorize=None):
    def _token_instructions(flow_name, prev_user):
        obj.echo(
            "There is an existing version of *%s* on Argo Workflows which was "
            "deployed by the user *%s*." % (flow_name, prev_user)
        )
        obj.echo(
            "To terminate this flow, you need to use the same production token that they used."
        )
        obj.echo(
            "Please reach out to them to get the token. Once you have it, call "
            "this command:"
        )
        obj.echo("    argo-workflows terminate --authorize MY_TOKEN RUN_ID", fg="green")
        obj.echo(
            'See "Organizing Results" at docs.metaflow.org for more information '
            "about production tokens."
        )

    validate_run_id(
        obj.workflow_name, obj.token_prefix, authorize, run_id, _token_instructions
    )

    # Trim prefix from run_id
    name = run_id[5:]
    obj.echo(
        "Terminating run *{run_id}* for {flow_name} ...".format(
            run_id=run_id, flow_name=obj.flow.name
        ),
        bold=True,
    )

    terminated = ArgoWorkflows.terminate(obj.flow.name, name)
    if terminated:
        obj.echo("\nRun terminated.")


@argo_workflows.command(help="List Argo Workflow templates for the flow.")
@click.option(
    "--all",
    default=False,
    is_flag=True,
    type=bool,
    help="list all Argo Workflow Templates (not just limited to this flow)",
)
@click.pass_obj
def list_workflow_templates(obj, all=None):
    for template_name in ArgoWorkflows.list_templates(obj.flow.name, all):
        obj.echo_always(template_name)


# Internal CLI command to run a heartbeat daemon in an Argo Workflows Daemon container.
@argo_workflows.command(hidden=True, help="start heartbeat process for a run")
@click.option("--run_id", required=True)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    default=None,
    help="Annotate all objects produced by Argo Workflows runs "
    "with the given tag. You can specify this option multiple "
    "times to attach multiple tags.",
)
@click.pass_obj
def heartbeat(obj, run_id, tags=None):
    # Try to register a run in case the start task has not taken care of it yet.
    obj.metadata.register_run_id(run_id, tags)
    # Start run heartbeat
    obj.metadata.start_run_heartbeat(obj.flow.name, run_id)
    # Keepalive loop
    while True:
        # Do not pollute daemon logs with anything unnecessary,
        # as they might be extremely long running.
        sleep(10)


def validate_run_id(
    workflow_name, token_prefix, authorize, run_id, instructions_fn=None
):
    """
    Validates that a run_id adheres to the Argo Workflows naming rules, and
    that it belongs to the current flow (accounting for project branch as well).
    """
    # Verify that user is trying to change an Argo workflow
    if not run_id.startswith("argo-"):
        raise RunIdMismatch(
            "Run IDs for flows executed through Argo Workflows begin with 'argo-'"
        )

    # Verify that run_id belongs to the Flow, and that branches match
    name = run_id[5:]
    workflow = ArgoWorkflows.get_execution(name)
    if workflow is None:
        raise MetaflowException("Could not find workflow *%s* on Argo Workflows" % name)

    owner, token, flow_name, branch_name, project_name = workflow

    # Verify we are operating on the correct Flow file compared to the running one.
    # Without this check, using --name could be used to run commands for arbitrary run_id's, disregarding the Flow in the file.
    if current.flow_name != flow_name:
        raise RunIdMismatch(
            "The workflow with the run_id *%s* belongs to the flow *%s*, not for the flow *%s*."
            % (run_id, flow_name, current.flow_name)
        )

    if project_name is not None:
        # Verify we are operating on the correct project.
        if current.get("project_name") != project_name:
            raise RunIdMismatch(
                "The workflow belongs to the project *%s*. "
                "Please use the project decorator or --name to target the correct project"
                % project_name
            )

        # Verify we are operating on the correct branch.
        if current.get("branch_name") != branch_name:
            raise RunIdMismatch(
                "The workflow belongs to the branch *%s*. "
                "Please use --branch, --production or --name to target the correct branch"
                % branch_name
            )

    # Verify that the production tokens match. We do not want to cache the token that was used though,
    # as the operations that require run_id validation can target runs not authored from the local environment
    if authorize is None:
        authorize = load_token(token_prefix)
    elif authorize.startswith("production:"):
        authorize = authorize[11:]

    if owner != get_username() and authorize != token:
        if instructions_fn:
            instructions_fn(flow_name=name, prev_user=owner)
        raise IncorrectProductionToken("Try again with the correct production token.")

    return True


def sanitize_for_argo(text):
    """
    Sanitizes a string so it does not contain characters that are not permitted in Argo Workflow resource names.
    """
    return (
        re.compile(r"^[^A-Za-z0-9]+")
        .sub("", text)
        .replace("_", "")
        .replace("@", "")
        .replace("+", "")
        .lower()
    )


def remap_status(status):
    """
    Group similar Argo Workflow statuses together in order to have similar output to step functions statuses.
    """
    STATUS_MAP = {"Error": "Failed"}
    return STATUS_MAP.get(status, status)



from __future__ import print_function

import importlib
import os
import re
import sys
import types

from collections import defaultdict, namedtuple

from importlib.abc import MetaPathFinder, Loader
from itertools import chain
from pathlib import Path

from metaflow.meta_files import read_info_file


#
# This file provides the support for Metaflow's extension mechanism which allows
# a Metaflow developer to extend metaflow by providing a package `metaflow_extensions`.
# Multiple such packages can be provided, and they will all be loaded into Metaflow in a
# way that is transparent to the user.
#
# NOTE: The conventions used here may change over time and this is an advanced feature.
#
# The general functionality provided here can be divided into three phases:
#   - Package discovery: in this part, packages that provide metaflow extensions
#     are discovered. This is contained in the `_get_extension_packages` function
#   - Integration with Metaflow: throughout the Metaflow code, extension points
#     are provided (they are given below in `_extension_points`). At those points,
#     the core Metaflow code will invoke functions to load the packages discovered
#     in the first phase. These functions are:
#       - get_modules: Returns all modules that are contributing to the extension
#         point; this is typically done first.
#       - load_module: Simple loading of a specific module
#       - load_globals: Utility function to load the globals from a module into
#         another globals()-like object
#       - alias_submodules: Determines the aliases for modules allowing metaflow.Z to alias
#         metaflow_extensions.X.Y.Z for example. This supports the __mf_promote_submodules__
#         construct as well as aliasing any modules present in the extension. This is
#         typically used in conjunction with lazy_load_aliases which takes care of actually
#         making the aliasing work lazily (ie: modules that are not already loaded are only
#         loaded on use).
#       - lazy_load_aliases: Adds loaders for all the module aliases produced by
#         alias_submodules for example
#       - multiload_globals: Convenience function to `load_globals` on all modules returned
#         by `get_modules`
#       - multiload_all: Convenience function to `load_globals` and
#         `lazy_load_aliases(alias_submodules()) on all modules returned by `get_modules`
#   - Packaging the extensions: when extensions need to be included in the code package,
#     this allows the extensions to be properly included (including potentially non .py
#     files). To support this:
#       - dump_module_info dumps information in the INFO file allowing packaging to work
#         in a Conda environment or a remote environment (it saves file paths, load order, etc)
#       - package_mfext_package: allows the packaging of a single extension
#       - package_mfext_all: packages all extensions
#
# The get_aliases_modules is used by Pylint to ignore some of the errors arising from
# aliasing packages

__all__ = (
    "load_module",
    "get_modules",
    "dump_module_info",
    "get_extensions_in_dir",
    "extension_info",
    "update_package_info",
    "get_aliased_modules",
    "package_mfext_package",
    "package_mfext_all",
    "load_globals",
    "alias_submodules",
    "EXT_PKG",
    "lazy_load_aliases",
    "multiload_globals",
    "multiload_all",
    "_ext_debug",
)

EXT_PKG = "metaflow_extensions"
EXT_CONFIG_REGEXP = re.compile(r"^mfextinit_[a-zA-Z0-9_-]+\.py$")
EXT_META_REGEXP = re.compile(r"^mfextmeta_[a-zA-Z0-9_-]+\.py$")
REQ_NAME = re.compile(r"^(([a-zA-Z0-9][a-zA-Z0-9._-]*[a-zA-Z0-9])|[a-zA-Z0-9]).*$")
EXT_EXCLUDE_SUFFIXES = [".pyc"]

# To get verbose messages, set METAFLOW_DEBUG_EXT to 1
DEBUG_EXT = os.environ.get("METAFLOW_DEBUG_EXT", False)

# This is extracted only from environment variable and here separately from
# metaflow_config to prevent nasty circular dependencies
EXTENSIONS_SEARCH_DIRS = os.environ.get("METAFLOW_EXTENSIONS_SEARCH_DIRS", "").split(
    os.pathsep
)

MFExtPackage = namedtuple("MFExtPackage", "package_name tl_package config_module")
MFExtModule = namedtuple("MFExtModule", "package_name tl_package module")


def load_module(module_name):
    _ext_debug("Loading module '%s'..." % module_name)
    return _attempt_load_module(module_name)


def get_modules(extension_point):
    modules_to_load = []
    if extension_point not in _extension_points:
        raise RuntimeError(
            "Metaflow extension point '%s' not supported" % extension_point
        )
    _ext_debug("Getting modules for extension point '%s'..." % extension_point)
    for pkg in _pkgs_per_extension_point.get(extension_point, []):
        _ext_debug(
            "    Found top-level '%s' from '%s'" % (pkg.tl_package, pkg.package_name)
        )
        m = _get_extension_config(
            pkg.package_name, pkg.tl_package, extension_point, pkg.config_module
        )
        if m:
            modules_to_load.append(m)
    _ext_debug("    Loaded %s" % str(modules_to_load))
    return modules_to_load


def dump_module_info(all_packages=None, pkgs_per_extension_point=None):
    if all_packages is None:
        all_packages = _all_packages
    if pkgs_per_extension_point is None:
        pkgs_per_extension_point = _pkgs_per_extension_point

    _filter_files_all(all_packages)
    sanitized_all_packages = dict()
    # Strip out root_paths (we don't need it and no need to expose user's dir structure)
    for k, v in all_packages.items():
        sanitized_all_packages[k] = {
            "root_paths": None,
            "meta_module": v["meta_module"],
            "files": v["files"],
            "version": v["version"],
            "package_version": v.get("package_version", "<unk>"),
            "extension_name": v.get("extension_name", "<unk>"),
        }
    return "ext_info", [sanitized_all_packages, pkgs_per_extension_point]


def get_extensions_in_dir(d):
    return _get_extension_packages(ignore_info_file=True, restrict_to_directories=[d])


def extension_info(packages=None):
    if packages is None:
        packages = _all_packages
    # Returns information about installed extensions so it it can be stored in
    # _graph_info.
    return {
        "installed": {
            k: {
                "dist_version": v["version"],
                "package_version": v.get("package_version", "<unk>"),
                "extension_name": v.get("extension_name", "<unk>"),
            }
            for k, v in packages.items()
        },
    }


def update_package_info(pkg_to_update=None, package_name=None, **kwargs):
    pkg = None
    if pkg_to_update:
        pkg = pkg_to_update
    elif package_name:
        pkg = _all_packages.get(package_name)
    for k, v in kwargs.items():
        if k in pkg:
            raise ValueError(
                "Trying to overwrite existing key '%s' for package %s" % (k, str(pkg))
            )
        pkg[k] = v
    return pkg


def get_aliased_modules():
    return _aliased_modules


def package_mfext_package(package_name):
    from metaflow.util import to_unicode

    _ext_debug("Packaging '%s'" % package_name)
    pkg_info = _all_packages.get(package_name, None)
    _filter_files_package(pkg_info)
    if pkg_info and pkg_info.get("root_paths", None):
        single_path = len(pkg_info["root_paths"]) == 1
        for p in pkg_info["root_paths"]:
            root_path = to_unicode(p)
            for f in pkg_info["files"]:
                f_unicode = to_unicode(f)
                fp = os.path.join(root_path, f_unicode)
                if single_path or os.path.isfile(fp):
                    _ext_debug("    Adding '%s'" % fp)
                    yield fp, os.path.join(EXT_PKG, f_unicode)


def package_mfext_all():
    # When packaging extensions, we always add a __init__.py to make
    # the packaged metaflow_extensions directory "self-contained" so that
    # python doesn't go and search other parts of the system for more
    # metaflow_extensions.
    yield os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "_empty_file.py"
    ), os.path.join(EXT_PKG, "__init__.py")

    for p in _all_packages:
        for path_tuple in package_mfext_package(p):
            yield path_tuple


def package_mfext_all_descriptions():
    return _all_packages


def load_globals(module, dst_globals, extra_indent=False):
    if extra_indent:
        extra_indent = "    "
    else:
        extra_indent = ""
    _ext_debug("%sLoading globals from '%s'" % (extra_indent, module.__name__))
    for n, o in module.__dict__.items():
        if not n.startswith("__") and not isinstance(o, types.ModuleType):
            _ext_debug("%s    Importing '%s'" % (extra_indent, n))
            dst_globals[n] = o


def alias_submodules(module, tl_package, extension_point, extra_indent=False):
    if extra_indent:
        extra_indent = "    "
    else:
        extra_indent = ""
    lazy_load_custom_modules = {}

    _ext_debug("%sAliasing submodules for '%s'" % (extra_indent, module.__name__))

    addl_modules = module.__dict__.get("__mf_promote_submodules__")
    if addl_modules:
        # We make an alias for these modules which the extension author wants to
        # expose but since it may not already be loaded, we don't load it either

        # TODO: This does not properly work for multiple packages that overwrite
        # their submodule for example if EXT_PKG.X.datatools.Y is provided
        # by two packages. For now, don't do this.
        if extension_point is not None:
            lazy_load_custom_modules.update(
                {
                    "metaflow.%s.%s"
                    % (extension_point, k): "%s.%s.%s.%s"
                    % (EXT_PKG, tl_package, extension_point, k)
                    for k in addl_modules
                }
            )
        else:
            # Top-level "metaflow" overrides
            lazy_load_custom_modules.update(
                {
                    "metaflow.%s" % k: "%s.%s.%s" % (EXT_PKG, tl_package, k)
                    for k in addl_modules
                }
            )
        if lazy_load_custom_modules:
            _ext_debug(
                "%s    Found explicit promotions in __mf_promote_submodules__: %s"
                % (extra_indent, str(list(lazy_load_custom_modules.keys())))
            )
    for n, o in module.__dict__.items():
        if (
            isinstance(o, types.ModuleType)
            and o.__package__
            and o.__package__.startswith("%s.%s" % (EXT_PKG, tl_package))
        ):
            # NOTE: The condition above prohibits loading across tl_packages. We
            # can relax if needed but may not be a great idea.
            if extension_point is not None:
                lazy_load_custom_modules["metaflow.%s.%s" % (extension_point, n)] = o
            else:
                lazy_load_custom_modules["metaflow.%s" % n] = o
    _ext_debug(
        "%s    Will create the following module aliases: %s"
        % (extra_indent, str(list(lazy_load_custom_modules.keys())))
    )
    _aliased_modules.extend(lazy_load_custom_modules.keys())
    return lazy_load_custom_modules


def lazy_load_aliases(aliases):
    if aliases:
        sys.meta_path = [_LazyFinder(aliases)] + sys.meta_path


def multiload_globals(modules, dst_globals):
    for m in modules:
        load_globals(m.module, dst_globals, extra_indent=True)


def multiload_all(modules, extension_point, dst_globals):
    for m in modules:
        # Note that we load aliases separately (as opposed to in one fell swoop) so
        # modules loaded later in `modules` can depend on them
        lazy_load_aliases(
            alias_submodules(m.module, m.tl_package, extension_point, extra_indent=True)
        )
        load_globals(m.module, dst_globals)


_py_ver = sys.version_info[:2]
_aliased_modules = []

import importlib.util

if _py_ver >= (3, 8):
    from importlib import metadata
elif _py_ver >= (3, 7):
    from metaflow._vendor.v3_7 import importlib_metadata as metadata
else:
    from metaflow._vendor.v3_6 import importlib_metadata as metadata

# Extension points are the directories that can be present in a EXT_PKG to
# contribute to that extension point. For example, if you have
# metaflow_extensions/X/plugins, your extension contributes to the plugins
# extension point.
# IMPORTANT: More specific paths must appear FIRST (before any less specific one). For
# efficiency, put the less specific ones directly under more specific ones.
_extension_points = [
    "plugins.env_escape",
    "plugins.cards",
    "plugins.datatools",
    "plugins",
    "config",
    "exceptions",
    "toplevel",
    "cmd",
    "alias",
]


def _ext_debug(*args, **kwargs):
    if DEBUG_EXT:
        init_str = "%s:" % EXT_PKG
        kwargs["file"] = sys.stderr
        print(init_str, *args, **kwargs)


def _get_extension_packages(ignore_info_file=False, restrict_to_directories=None):
    # If we have an INFO file with the appropriate information (if running from a saved
    # code package for example), we use that directly
    # Pre-compute on _extension_points
    info_content = read_info_file()
    if not ignore_info_file and info_content:
        all_pkg, ext_to_pkg = info_content.get("ext_info", (None, None))
        if all_pkg is not None and ext_to_pkg is not None:
            _ext_debug("Loading pre-computed information from INFO file")
            # We need to properly convert stuff in ext_to_pkg
            for k, v in ext_to_pkg.items():
                v = [MFExtPackage(*d) for d in v]
                ext_to_pkg[k] = v
            return all_pkg, ext_to_pkg

    # Late import to prevent some circular nastiness
    if restrict_to_directories is None and EXTENSIONS_SEARCH_DIRS != [""]:
        restrict_to_directories = EXTENSIONS_SEARCH_DIRS

    # Check if we even have extensions
    try:
        extensions_module = importlib.import_module(EXT_PKG)
    except ImportError as e:
        # e.name is set to the name of the package that fails to load
        # so don't error ONLY IF the error is importing this module (but do
        # error if there is a transitive import error)
        if not (isinstance(e, ModuleNotFoundError) and e.name == EXT_PKG):
            raise
        return {}, {}

    if restrict_to_directories:
        restrict_to_directories = [
            Path(p).resolve().as_posix() for p in restrict_to_directories
        ]

    # There are two "types" of packages:
    #   - those installed on the system (distributions)
    #   - those present in the PYTHONPATH
    # We have more information on distributions (including dependencies) and more
    # effective ways to get file information from them (they include the full list of
    # files installed) so we treat them separately from packages purely in PYTHONPATH.
    # They are also the more likely way that users will have extensions present, so
    # we optimize for that case.

    # At this point, we look at all the paths and create a set. As we find distributions
    # that match it, we will remove from the set and then will be left with any
    # PYTHONPATH "packages"
    all_paths = set(Path(p).resolve().as_posix() for p in extensions_module.__path__)
    _ext_debug("Found packages present at %s" % str(all_paths))
    if restrict_to_directories:
        _ext_debug(
            "Processed packages will be restricted to %s" % str(restrict_to_directories)
        )

    list_ext_points = [x.split(".") for x in _extension_points]
    init_ext_points = [x[0] for x in list_ext_points]

    # NOTE: For distribution packages, we will rely on requirements to determine the
    # load order of extensions: if distribution A and B both provide EXT_PKG and
    # distribution A depends on B then when returning modules in `get_modules`, we will
    # first return B and THEN A. We may want
    # other ways of specifying "load me after this if it exists" without depending on
    # the package. One way would be to rely on the description and have that info there.
    # Not sure of the use, though, so maybe we can skip for now.

    # Key: distribution name/package path
    # Value: Dict containing:
    #   root_paths: The root path for all the files in this package. Can be a list in
    #               some rare cases
    #   meta_module: The module to the meta file (if any) that contains information about
    #     how to package this extension (suffixes to include/exclude)
    #   files: The list of files to be included (or considered for inclusion) when
    #     packaging this extension
    mf_ext_packages = dict()

    # Key: extension point (one of _extension_point)
    # Value: another dictionary with
    #   Key: distribution name/full path to package
    #   Value: another dictionary with
    #    Key: Top-level package name (so in metaflow_extensions.X...., the X)
    #    Value: MFExtPackage
    extension_points_to_pkg = defaultdict(dict)

    # Key: string: configuration file for a package
    # Value: list: packages that this configuration file is present in
    config_to_pkg = defaultdict(list)
    # Same as config_to_pkg for meta files
    meta_to_pkg = defaultdict(list)

    # 1st step: look for distributions (the common case)
    for dist in metadata.distributions():
        if any(
            [pkg == EXT_PKG for pkg in (dist.read_text("top_level.txt") or "").split()]
        ):
            # In all cases (whether duplicate package or not), we remove the package
            # from the list of locations to look in.
            # This is not 100% accurate because it is possible that at the same
            # location there is a package and a non-package, but this is extremely
            # unlikely so we are going to ignore this case.
            dist_root = dist.locate_file(EXT_PKG).resolve().as_posix()
            all_paths.discard(dist_root)
            dist_name = dist.metadata["Name"]
            dist_version = dist.metadata["Version"]
            if restrict_to_directories:
                parent_dirs = list(
                    p.as_posix() for p in Path(dist_root).resolve().parents
                )
                if all(p not in parent_dirs for p in restrict_to_directories):
                    _ext_debug(
                        "Ignoring package at %s as it is not in the considered directories"
                        % dist_root
                    )
                    continue
            if dist_name in mf_ext_packages:
                _ext_debug(
                    "Ignoring duplicate package '%s' (duplicate paths in sys.path? (%s))"
                    % (dist_name, str(sys.path))
                )
                continue
            _ext_debug(
                "Found extension package '%s' at '%s'..." % (dist_name, dist_root)
            )

            files_to_include = []
            meta_module = None

            # At this point, we check to see what extension points this package
            # contributes to. This is to enable multiple namespace packages to contribute
            # to the same extension point (for example, you may have multiple packages
            # that have plugins)
            for f in dist.files:
                parts = list(f.parts)

                if len(parts) > 1 and parts[0] == EXT_PKG:
                    # Ensure that we don't have a __init__.py to force this package to
                    # be a NS package
                    if parts[1] == "__init__.py":
                        raise RuntimeError(
                            "Package '%s' providing '%s' is not an implicit namespace "
                            "package as required" % (dist_name, EXT_PKG)
                        )

                    # Record the file as a candidate for inclusion when packaging if
                    # needed
                    if not any(
                        parts[-1].endswith(suffix) for suffix in EXT_EXCLUDE_SUFFIXES
                    ):
                        files_to_include.append(os.path.join(*parts[1:]))

                    if parts[1] in init_ext_points:
                        # This is most likely a problem as we need an intermediate
                        # "identifier"
                        raise RuntimeError(
                            "Package '%s' should conform to '%s.X.%s' and not '%s.%s' where "
                            "X is your organization's name for example"
                            % (
                                dist_name,
                                EXT_PKG,
                                parts[1],
                                EXT_PKG,
                                parts[1],
                            )
                        )

                    # Check for any metadata; we can only have one metadata per
                    # distribution at most
                    if EXT_META_REGEXP.match(parts[1]) is not None:
                        potential_meta_module = ".".join([EXT_PKG, parts[1][:-3]])
                        if meta_module:
                            raise RuntimeError(
                                "Package '%s' defines more than one meta configuration: "
                                "'%s' and '%s' (at least)"
                                % (
                                    dist_name,
                                    meta_module,
                                    potential_meta_module,
                                )
                            )
                        meta_module = potential_meta_module
                        _ext_debug(
                            "Found meta '%s' for '%s'" % (meta_module, dist_full_name)
                        )
                        meta_to_pkg[meta_module].append(dist_full_name)

                if len(parts) > 3 and parts[0] == EXT_PKG:
                    # We go over _extension_points *in order* to make sure we get more
                    # specific paths first

                    # To give useful errors in case multiple top-level packages in
                    # one package
                    dist_full_name = "%s[%s]" % (dist_name, parts[1])
                    for idx, ext_list in enumerate(list_ext_points):
                        if (
                            len(parts) > len(ext_list) + 2
                            and parts[2 : 2 + len(ext_list)] == ext_list
                        ):
                            # Check if this is an "init" file
                            config_module = None

                            if len(parts) == len(ext_list) + 3 and (
                                EXT_CONFIG_REGEXP.match(parts[-1]) is not None
                                or parts[-1] == "__init__.py"
                            ):
                                parts[-1] = parts[-1][:-3]  # Remove the .py
                                config_module = ".".join(parts)

                                config_to_pkg[config_module].append(dist_full_name)
                            cur_pkg = (
                                extension_points_to_pkg[_extension_points[idx]]
                                .setdefault(dist_name, {})
                                .get(parts[1])
                            )
                            if cur_pkg is not None:
                                if (
                                    config_module is not None
                                    and cur_pkg.config_module is not None
                                ):
                                    raise RuntimeError(
                                        "Package '%s' defines more than one "
                                        "configuration file for '%s': '%s' and '%s'"
                                        % (
                                            dist_full_name,
                                            _extension_points[idx],
                                            config_module,
                                            cur_pkg.config_module,
                                        )
                                    )
                                if config_module is not None:
                                    _ext_debug(
                                        "    Top-level '%s' found config file '%s'"
                                        % (parts[1], config_module)
                                    )
                                    extension_points_to_pkg[_extension_points[idx]][
                                        dist_name
                                    ][parts[1]] = MFExtPackage(
                                        package_name=dist_name,
                                        tl_package=parts[1],
                                        config_module=config_module,
                                    )
                            else:
                                _ext_debug(
                                    "    Top-level '%s' extends '%s' with config '%s'"
                                    % (parts[1], _extension_points[idx], config_module)
                                )
                                extension_points_to_pkg[_extension_points[idx]][
                                    dist_name
                                ][parts[1]] = MFExtPackage(
                                    package_name=dist_name,
                                    tl_package=parts[1],
                                    config_module=config_module,
                                )
                            break
            mf_ext_packages[dist_name] = {
                "root_paths": [dist_root],
                "meta_module": meta_module,
                "files": files_to_include,
                "version": dist_version,
            }
    # At this point, we have all the packages that contribute to EXT_PKG,
    # we now check to see if there is an order to respect based on dependencies. We will
    # return an ordered list that respects that order and is ordered alphabetically in
    # case of ties. We do not do any checks because we rely on pip to have done those.
    # Basically topological sort based on dependencies.
    pkg_to_reqs_count = {}
    req_to_dep = {}
    for pkg_name in mf_ext_packages:
        req_count = 0
        req_pkgs = [
            REQ_NAME.match(x).group(1) for x in metadata.requires(pkg_name) or []
        ]
        for req_pkg in req_pkgs:
            if req_pkg in mf_ext_packages:
                req_count += 1
                req_to_dep.setdefault(req_pkg, []).append(pkg_name)
        pkg_to_reqs_count[pkg_name] = req_count

    # Find roots
    mf_pkg_list = []
    to_process = []
    for pkg_name, count in pkg_to_reqs_count.items():
        if count == 0:
            to_process.append(pkg_name)

    # Add them in alphabetical order
    to_process.sort()
    mf_pkg_list.extend(to_process)
    # Find rest topologically
    while to_process:
        next_round = []
        for pkg_name in to_process:
            del pkg_to_reqs_count[pkg_name]
            for dep in req_to_dep.get(pkg_name, []):
                cur_req_count = pkg_to_reqs_count[dep]
                if cur_req_count == 1:
                    next_round.append(dep)
                else:
                    pkg_to_reqs_count[dep] = cur_req_count - 1
        # Add those in alphabetical order
        next_round.sort()
        mf_pkg_list.extend(next_round)
        to_process = next_round

    # Check that we got them all
    if len(pkg_to_reqs_count) > 0:
        raise RuntimeError(
            "Unresolved dependencies in '%s': %s"
            % (EXT_PKG, ", and ".join("'%s'" % p for p in pkg_to_reqs_count))
        )

    _ext_debug("'%s' distributions order is %s" % (EXT_PKG, str(mf_pkg_list)))

    # We check if we have any additional packages that were not yet installed that
    # we need to use. We always put them *last* in the load order and put them
    # alphabetically.
    all_paths_list = list(all_paths)
    all_paths_list.sort()

    # This block of code is the equivalent of the one above for distributions except
    # for PYTHONPATH packages. The functionality is identical, but it looks a little
    # different because we construct the file list instead of having it nicely provided
    # to us.
    package_name_to_path = dict()
    if len(all_paths_list) > 0:
        _ext_debug("Non installed packages present at %s" % str(all_paths))
        for package_count, package_path in enumerate(all_paths_list):
            if restrict_to_directories:
                parent_dirs = list(
                    p.as_posix() for p in Path(package_path).resolve().parents
                )
                if all(p not in parent_dirs for p in restrict_to_directories):
                    _ext_debug(
                        "Ignoring non-installed package at %s as it is not in "
                        "the considered directories" % package_path
                    )
                    continue
            # We give an alternate name for the visible package name. It is
            # not exposed to the end user but used to refer to the package, and it
            # doesn't provide much additional information to have the full path
            # particularly when it is on a remote machine.
            # We keep a temporary mapping around for error messages while loading for
            # the first time.
            package_name = "_pythonpath_%d" % package_count
            _ext_debug(
                "Walking path %s (package name %s)" % (package_path, package_name)
            )
            package_name_to_path[package_name] = package_path
            base_depth = len(package_path.split("/"))
            files_to_include = []
            meta_module = None
            for root, dirs, files in os.walk(package_path):
                parts = root.split("/")
                cur_depth = len(parts)
                # relative_root strips out metaflow_extensions
                relative_root = "/".join(parts[base_depth:])
                relative_module = ".".join(parts[base_depth - 1 :])
                files_to_include.extend(
                    [
                        "/".join([relative_root, f]) if relative_root else f
                        for f in files
                        if not any(
                            [f.endswith(suffix) for suffix in EXT_EXCLUDE_SUFFIXES]
                        )
                    ]
                )
                if cur_depth == base_depth:
                    if "__init__.py" in files:
                        raise RuntimeError(
                            "'%s' at '%s' is not an implicit namespace package as required"
                            % (EXT_PKG, root)
                        )
                    for d in dirs:
                        if d in init_ext_points:
                            raise RuntimeError(
                                "Package at '%s' should conform to' %s.X.%s' and not "
                                "'%s.%s' where X is your organization's name for example"
                                % (root, EXT_PKG, d, EXT_PKG, d)
                            )
                    # Check for meta files for this package
                    meta_files = [
                        x for x in map(EXT_META_REGEXP.match, files) if x is not None
                    ]
                    if meta_files:
                        # We should have one meta file at most
                        if len(meta_files) > 1:
                            raise RuntimeError(
                                "Package at '%s' defines more than one meta file: %s"
                                % (
                                    package_path,
                                    ", and ".join(
                                        ["'%s'" % x.group(0) for x in meta_files]
                                    ),
                                )
                            )
                        else:
                            meta_module = ".".join(
                                [relative_module, meta_files[0].group(0)[:-3]]
                            )

                elif cur_depth > base_depth + 1:
                    # We want at least a top-level name and something under
                    tl_name = parts[base_depth]
                    tl_fullname = "%s[%s]" % (package_path, tl_name)
                    prefix_match = parts[base_depth + 1 :]
                    for idx, ext_list in enumerate(list_ext_points):
                        if prefix_match == ext_list:
                            # We check to see if this is an actual extension point
                            # or if we just have a directory on the way to another
                            # extension point. To do this, we check to see if we have
                            # any files or directories that are *not* directly another
                            # extension point
                            skip_extension = len(files) == 0
                            if skip_extension:
                                next_dir_idx = len(list_ext_points[idx])
                                ok_subdirs = [
                                    list_ext_points[j][next_dir_idx]
                                    for j in range(0, idx)
                                    if len(list_ext_points[j]) > next_dir_idx
                                ]
                                skip_extension = set(dirs).issubset(set(ok_subdirs))

                            if skip_extension:
                                _ext_debug(
                                    "    Skipping '%s' as no files/directory of interest"
                                    % _extension_points[idx]
                                )
                                continue

                            # Check for any "init" files
                            init_files = [
                                x.group(0)
                                for x in map(EXT_CONFIG_REGEXP.match, files)
                                if x is not None
                            ]
                            if "__init__.py" in files:
                                init_files.append("__init__.py")

                            config_module = None
                            if len(init_files) > 1:
                                raise RuntimeError(
                                    "Package at '%s' defines more than one configuration "
                                    "file for '%s': %s"
                                    % (
                                        tl_fullname,
                                        ".".join(prefix_match),
                                        ", and ".join(["'%s'" % x for x in init_files]),
                                    )
                                )
                            elif len(init_files) == 1:
                                config_module = ".".join(
                                    [relative_module, init_files[0][:-3]]
                                )
                                config_to_pkg[config_module].append(tl_fullname)

                            d = extension_points_to_pkg[_extension_points[idx]][
                                package_name
                            ] = dict()
                            d[tl_name] = MFExtPackage(
                                package_name=package_name,
                                tl_package=tl_name,
                                config_module=config_module,
                            )
                            _ext_debug(
                                "    Extends '%s' with config '%s'"
                                % (_extension_points[idx], config_module)
                            )
            if files_to_include:
                mf_pkg_list.append(package_name)
                mf_ext_packages[package_name] = {
                    "root_paths": [package_path],
                    "meta_module": meta_module,
                    "files": files_to_include,
                    "version": "_local_",
                }
            else:
                _ext_debug("Skipping package as no files found (empty dir?)")

    # Sanity check that we only have one package per configuration file.
    # This prevents multiple packages from providing the same named configuration
    # file which would result in one overwriting the other if they are both installed.
    errors = []
    for m, packages in config_to_pkg.items():
        if len(packages) > 1:
            errors.append(
                "    Packages %s define the same configuration module '%s'"
                % (", and ".join(["'%s'" % p for p in packages]), m)
            )
    for m, packages in meta_to_pkg.items():
        if len(packages) > 1:
            errors.append(
                "    Packages %s define the same meta module '%s'"
                % (", and ".join(["'%s'" % p for p in packages]), m)
            )
    if errors:
        raise RuntimeError(
            "Conflicts in '%s' files:\n%s" % (EXT_PKG, "\n".join(errors))
        )

    extension_points_to_pkg.default_factory = None

    # We have the load order globally; we now figure it out per extension point.
    for k, v in extension_points_to_pkg.items():
        # v is a dict distributionName/packagePath -> (dict tl_name -> MFPackage)
        l = [v[pkg].values() for pkg in mf_pkg_list if pkg in v]
        # In the case of the plugins.cards extension we allow those packages
        # to be ns packages, so we only list the package once (in its first position).
        # In all other cases, we error out if we don't have a configuration file for the
        # package (either a __init__.py of an explicit mfextinit_*.py)
        final_list = []
        null_config_tl_package = set()
        for pkg in chain(*l):
            if pkg.config_module is None:
                if k == "plugins.cards":
                    # This is allowed here but we only keep one
                    if pkg.tl_package in null_config_tl_package:
                        continue
                    null_config_tl_package.add(pkg.tl_package)
                else:
                    package_path = package_name_to_path.get(pkg.package_name)
                    if package_path:
                        package_path = "at '%s'" % package_path
                    else:
                        package_path = "'%s'" % pkg.package_name
                    raise RuntimeError(
                        "Package %s does not define a configuration file for '%s'"
                        % (package_path, k)
                    )
            final_list.append(pkg)
        extension_points_to_pkg[k] = final_list
    return mf_ext_packages, extension_points_to_pkg


_all_packages, _pkgs_per_extension_point = _get_extension_packages()


def _attempt_load_module(module_name):
    try:
        extension_module = importlib.import_module(module_name)
    except ImportError as e:
        # e.name is set to the name of the package that fails to load
        # so don't error ONLY IF the error is importing this module (but do
        # error if there is a transitive import error)
        errored_names = [EXT_PKG]
        parts = module_name.split(".")
        for p in parts[1:]:
            errored_names.append("%s.%s" % (errored_names[-1], p))
        if not (isinstance(e, ModuleNotFoundError) and e.name in errored_names):
            print(
                "The following exception occurred while trying to load '%s' ('%s')"
                % (EXT_PKG, module_name)
            )
            raise
        _ext_debug("        Unknown error when loading '%s': %s" % (module_name, e))
        return None
    else:
        return extension_module


def _get_extension_config(distribution_name, tl_pkg, extension_point, config_module):
    if config_module is not None and not config_module.endswith("__init__"):
        module_name = config_module
        # file_path below will be /root/metaflow_extensions/X/Y/mfextinit_Z.py and
        # module name is metaflow_extensions.X.Y.mfextinit_Z so if we want to strip to
        # /root/metaflow_extensions, we need to remove this number of elements from the
        # filepath
        strip_from_filepath = len(module_name.split(".")) - 1
    else:
        module_name = ".".join([EXT_PKG, tl_pkg, extension_point])
        # file_path here will be /root/metaflow_extensions/X/Y/__init__.py BUT
        # module name is metaflow_extensions.X.Y so we have a 1 off compared to the
        # previous case
        strip_from_filepath = len(module_name.split("."))

    _ext_debug("        Attempting to load '%s'" % module_name)

    extension_module = _attempt_load_module(module_name)

    if extension_module:
        # We update the path to this module. This is useful if we need to package this
        # package again. Note that in most cases, packaging happens in the outermost
        # local python environment (non Conda and not remote) so we already have the
        # root_paths set when we are initially looking for metaflow_extensions package.
        # This code allows for packaging while running inside a Conda environment or
        # remotely where the root_paths has been changed since the initial packaging.
        # This currently does not happen much.
        if _all_packages[distribution_name]["root_paths"] is None:
            file_path = getattr(extension_module, "__file__")
            if file_path:
                # Common case where this is an actual init file (mfextinit_X.py or __init__.py)
                root_paths = ["/".join(file_path.split("/")[:-strip_from_filepath])]
            else:
                # Only used for plugins.cards where the package can be a NS package. In
                # this case, __path__ will have things like /root/metaflow_extensions/X/Y
                # and module name will be metaflow_extensions.X.Y
                root_paths = [
                    "/".join(p.split("/")[: -len(module_name.split(".")) + 1])
                    for p in extension_module.__path__
                ]

            _ext_debug("Package '%s' is rooted at %s" % (distribution_name, root_paths))
            _all_packages[distribution_name]["root_paths"] = root_paths

        return MFExtModule(
            package_name=distribution_name, tl_package=tl_pkg, module=extension_module
        )
    return None


def _filter_files_package(pkg):
    if pkg and pkg["root_paths"] and pkg["meta_module"]:
        meta_module = _attempt_load_module(pkg["meta_module"])
        if meta_module:
            include_suffixes = meta_module.__dict__.get("include_suffixes")
            exclude_suffixes = meta_module.__dict__.get("exclude_suffixes")

            # Behavior is as follows:
            #  - if nothing specified, include all files (so do nothing here)
            #  - if include_suffixes, only include those suffixes
            #  - if *not* include_suffixes but exclude_suffixes, include everything *except*
            #    files ending with that suffix
            if include_suffixes:
                new_files = [
                    f
                    for f in pkg["files"]
                    if any([f.endswith(suffix) for suffix in include_suffixes])
                ]
            elif exclude_suffixes:
                new_files = [
                    f
                    for f in pkg["files"]
                    if not any([f.endswith(suffix) for suffix in exclude_suffixes])
                ]
            else:
                new_files = pkg["files"]
            pkg["files"] = new_files


def _filter_files_all(all_packages):
    for p in all_packages.values():
        _filter_files_package(p)


class _AliasLoader(Loader):
    def __init__(self, alias, orig):
        self._alias = alias
        self._orig = orig

    def create_module(self, spec):
        _ext_debug(
            "Loading aliased module '%s' at '%s' " % (str(self._orig), spec.name)
        )
        if isinstance(self._orig, str):
            try:
                return importlib.import_module(self._orig)
            except ImportError:
                raise ImportError(
                    "No module found '%s' (aliasing '%s')" % (spec.name, self._orig)
                )
        elif isinstance(self._orig, types.ModuleType):
            # We are aliasing a module, so we just return that one
            return self._orig
        else:
            return super().create_module(spec)

    def exec_module(self, module):
        # Override the name to make it a bit nicer. We keep the old name so that
        # we can refer to it when we load submodules
        if not hasattr(module, "__orig_name__"):
            module.__orig_name__ = module.__name__
            module.__name__ = self._alias


class _OrigLoader(Loader):
    def __init__(
        self,
        fullname,
        orig_loader,
        previously_loaded_module=None,
        previously_loaded_parent_module=None,
    ):
        self._fullname = fullname
        self._orig_loader = orig_loader
        self._previously_loaded_module = previously_loaded_module
        self._previously_loaded_parent_module = previously_loaded_parent_module

    def create_module(self, spec):
        _ext_debug(
            "Loading original module '%s' (will be loaded at '%s'); spec is %s"
            % (spec.name, self._fullname, str(spec))
        )
        self._orig_name = spec.name
        return self._orig_loader.create_module(spec)

    def exec_module(self, module):
        try:
            # Perform all actions of the original loader
            self._orig_loader.exec_module(module)
        except BaseException:
            raise  # We re-raise it always; the `finally` clause will still restore things
        else:
            # It loaded, we move and rename appropriately
            module.__spec__.name = self._fullname
            module.__orig_name__ = module.__name__
            module.__name__ = self._fullname
            module.__package__ = module.__spec__.parent  # assumption since 3.6
            sys.modules[self._fullname] = module
            del sys.modules[self._orig_name]

        finally:
            # At this point, the original module is loaded with the original name. We
            # want to replace it with previously_loaded_module if it exists. We
            # also replace the parent properly
            if self._previously_loaded_module:
                sys.modules[self._orig_name] = self._previously_loaded_module
            if self._previously_loaded_parent_module:
                sys.modules[".".join(self._orig_name.split(".")[:-1])] = (
                    self._previously_loaded_parent_module
                )


class _LazyFinder(MetaPathFinder):
    # This _LazyFinder implements the Importer Protocol defined in PEP 302

    def __init__(self, handled):
        # Dictionary:
        # Key: name of the module to handle
        # Value:
        #   - A string: a pathspec to the module to load
        #   - A module: the module to load
        self._handled = handled if handled else {}

        # This is used to revert to regular loading when trying to load
        # the over-ridden module
        self._temp_excluded_prefix = set()

        # This is used to determine if we should be searching in _orig modules. Basically,
        # when a relative import is done from a module in _orig, we want to search in
        # the _orig "tree"
        self._orig_search_paths = set()

    def find_spec(self, fullname, path, target=None):
        # If we are trying to load a shadowed module (ending in ._orig), we don't
        # say we handle it
        # _ext_debug(
        #    "Looking for %s in %s with target %s" % (fullname, str(path), target)
        # )
        if any([fullname.startswith(e) for e in self._temp_excluded_prefix]):
            return None

        # If this is something we directly handle, return our loader
        if fullname in self._handled:
            return importlib.util.spec_from_loader(
                fullname, _AliasLoader(fullname, self._handled[fullname])
            )

        # For the first pass when we try to load a shadowed module, we send it back
        # without the ._orig and that will find the original spec of the module
        # Note that we handle mymodule._orig.orig_submodule as well as mymodule._orig.
        # Basically, the original module and any of the original submodules are
        # available under _orig.
        name_parts = fullname.split(".")
        try:
            orig_idx = name_parts.index("_orig")
        except ValueError:
            orig_idx = -1
        if orig_idx > -1 and ".".join(name_parts[:orig_idx]) in self._handled:
            orig_name = ".".join(name_parts[:orig_idx] + name_parts[orig_idx + 1 :])
            parent_name = None
            if orig_idx != len(name_parts) - 1:
                # We have a parent module under the _orig portion so for example, if
                # we load mymodule._orig.orig_submodule, our parent is mymodule._orig.
                # However, since mymodule is currently shadowed, we need to reset
                # the parent module properly. We know it is already loaded (since modules
                # are loaded hierarchically)
                parent_name = ".".join(
                    name_parts[:orig_idx] + name_parts[orig_idx + 1 : -1]
                )
            _ext_debug("Looking for original module '%s'" % orig_name)
            prefix = ".".join(name_parts[:orig_idx])
            self._temp_excluded_prefix.add(prefix)
            # We also have to remove the module temporarily while we look for the
            # new spec since otherwise it returns the spec of that loaded module.
            # module is also restored *after* we call `create_module` in the loader
            # otherwise it just returns None. We also swap out the parent module so that
            # the search can start from there.
            loaded_module = sys.modules.get(orig_name)
            if loaded_module:
                del sys.modules[orig_name]
            parent_module = sys.modules.get(parent_name) if parent_name else None
            if parent_module:
                sys.modules[parent_name] = sys.modules[".".join([parent_name, "_orig"])]

            # This finds the spec that would have existed had we not added all our
            # _LazyFinders
            spec = importlib.util.find_spec(orig_name)

            self._temp_excluded_prefix.remove(prefix)

            if not spec:
                return None

            if spec.submodule_search_locations:
                self._orig_search_paths.update(spec.submodule_search_locations)

            _ext_debug("Found original spec %s" % spec)

            # Change the spec
            spec.loader = _OrigLoader(
                fullname,
                spec.loader,
                loaded_module,
                parent_module,
            )

            return spec

        for p in path or []:
            if p in self._orig_search_paths:
                # We need to look in some of the "_orig" modules
                orig_override_name = ".".join(
                    name_parts[:-1] + ["_orig", name_parts[-1]]
                )
                _ext_debug(
                    "Looking for %s as an original module: searching for %s"
                    % (fullname, orig_override_name)
                )
                return importlib.util.find_spec(orig_override_name)
        if len(name_parts) > 1:
            # This checks for submodules of things we handle. We check for the most
            # specific submodule match and use that
            chop_idx = 1
            while chop_idx < len(name_parts):
                parent_name = ".".join(name_parts[:-chop_idx])
                if parent_name in self._handled:
                    orig = self._handled[parent_name]
                    if isinstance(orig, types.ModuleType):
                        orig_name = ".".join(
                            [orig.__orig_name__] + name_parts[-chop_idx:]
                        )
                    else:
                        orig_name = ".".join([orig] + name_parts[-chop_idx:])
                    return importlib.util.spec_from_loader(
                        fullname, _AliasLoader(fullname, orig_name)
                    )
                chop_idx += 1
        return None



import importlib
import traceback

from metaflow.metaflow_config_funcs import from_conf

from . import _ext_debug, get_modules

# This file is similar in functionality to the cmd.py file. Please refer to that
# one for more information on what the functions do.


def process_integration_aliases(module_globals):
    _resolve_relative_paths(module_globals)

    all_aliases = _get_ext_aliases(module_globals)
    all_aliases_dict = {}

    toggle_alias = []
    list_of_aliases = from_conf("ENABLED_INTEGRATION_ALIAS")

    try:
        modules_to_import = get_modules("alias")
        # This is like multiload_all but we load globals independently since we just care
        # about the TOGGLE and ENABLED values
        for m in modules_to_import:
            for n, o in m.module.__dict__.items():
                if n == "TOGGLE_INTEGRATION_ALIAS":
                    toggle_alias.extend(o)
                elif n == "ENABLED_INTEGRATION_ALIAS":
                    list_of_aliases = o
            _resolve_relative_paths(m.module.__dict__)
            all_aliases.extend(_get_ext_aliases(m.module.__dict__))
    except Exception as e:
        _ext_debug(
            "\tWARNING: ignoring all integration aliases due to error during import: %s"
            % e
        )
        print(
            "WARNING: Integration aliases did not load -- ignoring all of them which "
            "may not be what you want: %s" % e
        )
        traceback.print_exc()

    # At this point, we have _all_aliases populated with all the tuples
    # (name, module_class) from all the aliases in all the extensions (if any)
    # We build a dictionary taking the latest presence for each name (so plugins
    # override metaflow core)
    for name, obj_path in all_aliases:
        _ext_debug("    Adding integration alias '%s' from '%s'" % (name, obj_path))
        all_aliases_dict[name] = obj_path

    # Resolve the ENABLED_INTEGRATION_ALIAS variable. The rules are the following:
    #  - if ENABLED_INTEGRATION_ALIAS is non None, it means it was either set directly
    #    by the user in a configuration file, on the command line or by an extension.
    #    In that case we honor those wishes and completely ignore the extensions' toggles.
    #  - if ENABLED_INTEGRATION_ALIAS is None, we populate it with everything included
    #    here and in all the extensions and use the TOGGLE_ list to produce the final list.
    # The rationale behind this is to support both a configuration option where the
    # aliases enabled are explicitly listed (typical in a lot of software) but also to
    # support a "configuration-less" version where the installation of the extensions
    # determines what is activated.
    if list_of_aliases is None:
        list_of_aliases = list(all_aliases_dict) + toggle_alias

    _ext_debug("    Resolving metaflow integration aliases")
    _ext_debug("        Raw list is: %s" % str(list_of_aliases))

    set_of_aliases = set()
    for p in list_of_aliases:
        if p.startswith("-"):
            set_of_aliases.discard(p[1:])
        elif p.startswith("+"):
            set_of_aliases.add(p[1:])
        else:
            set_of_aliases.add(p)
    _ext_debug("        Resolved list is: %s" % str(set_of_aliases))

    for name in set_of_aliases:
        obj_path = all_aliases_dict.get(name, None)
        if obj_path is None:
            raise ValueError(
                "Configuration requested integration alias '%s' but no such alias "
                "is available" % name
            )
        path, obj_name = obj_path.rsplit(".", 1)
        try:
            alias_module = importlib.import_module(path)
        except ImportError:
            raise ValueError(
                "Cannot locate integration alias '%s' at '%s'" % (name, path)
            )

        obj = getattr(alias_module, obj_name, None)
        if obj is None:
            raise ValueError(
                "Cannot locate '%s' object for integration alias at '%s'"
                % (obj_name, path)
            )
        _ext_debug("        Added integration alias '%s' from '%s'" % (name, obj_path))
        module_globals[name] = obj


def _get_ext_aliases(module_globals):
    return module_globals.get("ALIASES_DESC", [])


def _set_ext_aliases(module_globals, value):
    module_globals["ALIASES_DESC"] = value


def _resolve_relative_paths(module_globals):
    # We want to modify all the relevant lists so that the relative paths
    # are made fully qualified paths for the modules
    pkg_path = module_globals["__package__"]
    pkg_components = pkg_path.split(".")

    def resolve_path(class_path):
        # Converts a relative class_path to an absolute one considering that the
        # relative class_path is present in a package pkg_path
        if class_path[0] == ".":
            i = 1
            # Check for multiple "." at the start of the class_path
            while class_path[i] == ".":
                i += 1
            if i > len(pkg_components):
                raise ValueError(
                    "Path '%s' exits out of Metaflow module at %s"
                    % (class_path, pkg_path)
                )
            return (
                ".".join(pkg_components[: -i + 1] if i > 1 else pkg_components)
                + class_path[i - 1 :]
            )
        return class_path

    _set_ext_aliases(
        module_globals,
        list(
            map(lambda p: (p[0], resolve_path(p[1])), _get_ext_aliases(module_globals))
        ),
    )



import sys
from hashlib import md5


def generate_input_paths(step_name, timestamp, input_paths, split_cardinality):
    # => run_id/step/:foo,bar
    run_id = input_paths.split("/")[0]
    foreach_base_id = "{}-{}-{}".format(step_name, timestamp, input_paths)

    ids = [_generate_task_id(foreach_base_id, i) for i in range(int(split_cardinality))]
    return "{}/{}/:{}".format(run_id, step_name, ",".join(ids))


def _generate_task_id(base, idx):
    # For foreach splits generate the expected input-paths based on split_cardinality and base_id.
    # newline required at the end due to 'echo' appending one in the shell side task_id creation.
    task_str = "%s-%s\n" % (base, idx)
    hash = md5(task_str.encode("utf-8")).hexdigest()[-8:]
    return "t-" + hash


if __name__ == "__main__":
    print(generate_input_paths(sys.argv[1], sys.argv[2], sys.argv[3], sys.argv[4]))



import importlib
import traceback

from metaflow.metaflow_config_funcs import from_conf

from . import _ext_debug, get_modules

_all_cmds = []
_all_cmds_dict = {}

# Set ENABLED_ and _TOGGLE_ variables for commands
ENABLED_CMD = from_conf("ENABLED_CMD")
_TOGGLE_CMD = []

# This file is identical in functionality to the plugins.py file. Please refer to that
# one for more information on what the functions do.


def process_cmds(module_globals):
    global _all_cmds, _all_cmds_dict, ENABLED_CMD, _TOGGLE_CMD

    _resolve_relative_paths(module_globals)

    _all_cmds = _get_ext_cmds(module_globals)

    try:
        modules_to_import = get_modules("cmd")
        # This is like multiload_all but we load globals independently since we just care
        # about the TOGGLE and ENABLED values
        for m in modules_to_import:
            for n, o in m.module.__dict__.items():
                if n == "TOGGLE_CMD":
                    _TOGGLE_CMD.extend(o)
                elif n == "ENABLED_CMD":
                    ENABLED_CMD = o
            _resolve_relative_paths(m.module.__dict__)
            _all_cmds.extend(_get_ext_cmds(m.module.__dict__))
    except Exception as e:
        _ext_debug("\tWARNING: ignoring all cmds due to error during import: %s" % e)
        print(
            "WARNING: Cmds did not load -- ignoring all of them which may not "
            "be what you want: %s" % e
        )
        traceback.print_exc()

    # At this point, we have _all_cmds populated with all the tuples
    # (name, module_class) from all the cmds in all the extensions (if any)
    # We build a dictionary taking the latest presence for each name (so plugins
    # override metaflow core)
    for name, class_path in _all_cmds:
        _ext_debug("    Adding command '%s' from '%s'" % (name, class_path))
        _all_cmds_dict[name] = class_path

    # Resolve the ENABLED_CMD variable. The rules are the following:
    #  - if ENABLED_CMD is non None, it means it was either set directly by the user
    #    in a configuration file, on the command line or by an extension. In that case
    #    we honor those wishes and completely ignore the extensions' toggles.
    #  - if ENABLED_CMD is None, we populate it with everything included here and in
    #    all the extensions and use the TOGGLE_ list to produce the final list.
    # The rationale behind this is to support both a configuration option where the
    # cmds enabled are explicitly listed (typical in a lot of software) but also to
    # support a "configuration-less" version where the installation of the extensions
    # determines what is activated.
    if ENABLED_CMD is None:
        ENABLED_CMD = list(_all_cmds_dict) + _TOGGLE_CMD


def resolve_cmds():
    _ext_debug("    Resolving metaflow commands")
    list_of_cmds = ENABLED_CMD
    _ext_debug("        Raw list is: %s" % str(list_of_cmds))

    set_of_commands = set()
    for p in list_of_cmds:
        if p.startswith("-"):
            set_of_commands.discard(p[1:])
        elif p.startswith("+"):
            set_of_commands.add(p[1:])
        else:
            set_of_commands.add(p)
    _ext_debug("        Resolved list is: %s" % str(set_of_commands))

    to_return = []

    for name in set_of_commands:
        class_path = _all_cmds_dict.get(name, None)
        if class_path is None:
            raise ValueError(
                "Configuration requested command '%s' but no such command is available"
                % name
            )
        path, cls_name = class_path.rsplit(".", 1)
        try:
            cmd_module = importlib.import_module(path)
        except ImportError:
            raise ValueError("Cannot locate command '%s' at '%s'" % (name, path))

        cls = getattr(cmd_module, cls_name, None)
        if cls is None:
            raise ValueError(
                "Cannot locate '%s' class for command at '%s'" % (cls_name, path)
            )
        all_cmds = list(cls.commands)
        if len(all_cmds) > 1:
            raise ValueError("%s defines more than one command -- use a group" % path)
        if all_cmds[0] != name:
            raise ValueError(
                "%s: expected name to be '%s' but got '%s' instead"
                % (path, name, all_cmds[0])
            )
        to_return.append(cls)
        _ext_debug("        Added command '%s' from '%s'" % (name, class_path))

    return to_return


def _get_ext_cmds(module_globals):
    return module_globals.get("CMDS_DESC", [])


def _set_ext_cmds(module_globals, value):
    module_globals["CMDS_DESC"] = value


def _resolve_relative_paths(module_globals):
    # We want to modify all the relevant lists so that the relative paths
    # are made fully qualified paths for the modules
    pkg_path = module_globals["__package__"]
    pkg_components = pkg_path.split(".")

    def resolve_path(class_path):
        # Converts a relative class_path to an absolute one considering that the
        # relative class_path is present in a package pkg_path
        if class_path[0] == ".":
            i = 1
            # Check for multiple "." at the start of the class_path
            while class_path[i] == ".":
                i += 1
            if i > len(pkg_components):
                raise ValueError(
                    "Path '%s' exits out of Metaflow module at %s"
                    % (class_path, pkg_path)
                )
            return (
                ".".join(pkg_components[: -i + 1] if i > 1 else pkg_components)
                + class_path[i - 1 :]
            )
        return class_path

    _set_ext_cmds(
        module_globals,
        list(map(lambda p: (p[0], resolve_path(p[1])), _get_ext_cmds(module_globals))),
    )



from metaflow._vendor import click
from hashlib import sha1
from metaflow.package import MetaflowPackage


@click.group()
def cli():
    pass


@cli.group(help="Commands related to code packages.")
@click.option(
    "--timeout", default=60, help="Timeout for package operations in seconds."
)
@click.pass_obj
def package(obj, timeout):
    # Prepare the package before any of the sub-commands are invoked.
    # We explicitly will *not* upload it to the datastore.
    obj.package = MetaflowPackage(
        obj.flow,
        obj.environment,
        obj.echo,
        suffixes=obj.package_suffixes,
        flow_datastore=None,
    )
    obj.package_op_timeout = timeout


@package.command(help="Output information about the code package.")
@click.pass_obj
def info(obj):
    obj.echo_always(obj.package.show())


@package.command(help="List all files included in the code package.")
@click.option(
    "--archive/--no-archive",
    default=False,
    help="If True, lists the file paths as present in the code package archive; "
    "otherwise, lists the files on your filesystem included in the code package",
    show_default=True,
)
@click.pass_obj
def list(obj, archive=False):
    _ = obj.package.blob_with_timeout(timeout=obj.package_op_timeout)
    # We now have all the information about the blob
    obj.echo(
        "Files included in the code package (change with --package-suffixes):",
        fg="magenta",
        bold=False,
    )
    if archive:
        obj.echo_always("\n".join(path for _, path in obj.package.path_tuples()))
    else:
        obj.echo_always("\n".join(path for path, _ in obj.package.path_tuples()))


@package.command(help="Save the current code package to a file.")
@click.argument("path")
@click.pass_obj
def save(obj, path):
    with open(path, "wb") as f:
        f.write(obj.package.blob())
    obj.echo(
        "Code package saved in *%s* with metadata: %s"
        % (path, obj.package.package_metadata),
        fg="magenta",
        bold=False,
    )
